/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btctl-audio.c,v 1.6 2002/11/25 15:24:13 kds Exp $

   Utility functions to read and parse the BT server config file. 

   Fixes:
   		Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#include <affix/config.h>

#include <sys/fcntl.h>
#include <sys/errno.h>
#include <sys/soundcard.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <termios.h>
#include <fcntl.h>

#include <affix/btcore.h>

#include "btctl.h"

/*
 * Audio stuff
 */

#define AUDIO_DEVICE	"/dev/dsp"
#define AUDIO_FILLSZ	4608
#define SPEED 8184
#define STEREO 0


int cmd_audio(struct btctl_command *cmd)
{
	int	fd, value, err = 0;

	fd = hci_open(btdev);
	if (fd == -1) {
		printf("Unable to open device %s: %s\n", btdev, strerror(errno));
		return -1;
	}

	switch (cmd->cmd) {
		case 0: {
				if (argv[argind] == NULL)
					value = 0x01; //set default
				else {
					err = sscanf(argv[argind], "%d", &value);
					if (err <= 0) {
						printf("input parameter error, using default value...\n");
						value = 0x01;
					}
				}

				err = HCI_EricssonSetSCODataPath(fd, value);
			} 
			break;
		case 1:
			{
				__u16	setting;
				int	mode = 0;

				if (argv[argind] == NULL) {
					err = HCI_ReadVoiceSetting(fd, &setting);
					if (err == 0)
						printf("Voice setting: %#03hx\n", setting);
				} else {
					err = sscanf(argv[argind], "%hx", &setting);
					if (err <= 0) {
						printf("input parameter error, using default value...\n");
						value = 0x060;
					}
					if (argv[argind+1]) { 
						if (strcmp(argv[argind+1], "in") == 0)
							mode |= AFFIX_AUDIO_IN;
						else if (strcmp(argv[argind+1], "out") == 0)
							mode |= AFFIX_AUDIO_OUT;
						else if (strcmp(argv[argind+1], "both") == 0)
							mode |= AFFIX_AUDIO_IN | AFFIX_AUDIO_OUT;
					}
					err = hci_set_audio(fd, mode, setting);
				}
			}
			break;
		case 2:
			{
				__u8	flow;

				if (argv[argind] == NULL) {
					err = HCI_ReadSCOFlowControl(fd, &flow);
					if (err == 0)
						printf("SCO Flow Control: %d\n", flow);
				} else {
					err = sscanf(argv[argind], "%d", &value);
					if (err <= 0) {
						printf("input parameter error, using default value...\n");
						value = 0;
					}
					err = HCI_WriteSCOFlowControl(fd, value);
				}
			}
			break;
	}
	if (err) {
		fprintf(stderr, "%s", hcierror(err));
		exit(1);
	}
	hci_close(fd);
	return 0;
}



static int audio_output(int fd, void const *buf, unsigned int len)
{
	char const *ptr = buf;
	int wrote;

	while (len) {
		wrote = write(fd, ptr, len);
		if (wrote == -1) {
			if (errno == EINTR)
				continue;
			else
				return -1;
		}

		ptr += wrote;
		len -= wrote;
	}
	return 0;
}

static int audio_buffer(int fd, void const *buf, unsigned int len)
{
	char const *ptr = buf;
	static char hold[AUDIO_FILLSZ];
	static unsigned int held;
	unsigned int left, grab;

	if (len == 0) {
		if (held) {
			memset(&hold[held], 0, &hold[AUDIO_FILLSZ] - &hold[held]);
			held = 0;

			return audio_output(fd, hold, AUDIO_FILLSZ);
		}

		return 0;
	}

	if (held == 0 && len == AUDIO_FILLSZ)
		return audio_output(fd, ptr, len);

	left = AUDIO_FILLSZ - held;

	while (len) {
		grab = len < left ? len : left;

		memcpy(&hold[held], ptr, grab);
		held += grab;
		left -= grab;

		ptr  += grab;
		len  -= grab;

		if (left == 0) {
			if (audio_output(fd, hold, AUDIO_FILLSZ) == -1)
				return -1;

			held = 0;
			left = AUDIO_FILLSZ;
		}
	}

	return 0;
}

# define audio_flush(fd)  audio_buffer((fd), 0, 0)

static int audio_init(int fd)
{
	int format, stereo, speed;

	format = AFMT_S16_LE;
	if (ioctl(fd, SNDCTL_DSP_SETFMT, &format) == -1) {
		perror("ioctl(SNDCTL_DSP_SETFMT)");
		return -1;
	}
	if (format != AFMT_S16_LE) {
		fprintf(stderr, "AFMT_S16_LE not available\n");
		return -1;
	}

#if 1
	stereo = STEREO;
	if (ioctl(fd, SNDCTL_DSP_STEREO, &stereo) == -1) {
		perror("ioctl(SNDCTL_DSP_STEREO)");
		return -1;
	}
	if (!stereo) {
		fprintf(stderr, "stereo not selected\n");
		//return -1;
	}
#endif
	speed = SPEED;
	if (ioctl(fd, SNDCTL_DSP_SPEED, &speed) == -1) {
		perror("ioctl(SNDCTL_DSP_SPEED)");
		return -1;
	}
	if (speed != SPEED) {
		fprintf(stderr, "sample speed SPEED not available (closest %u)\n", speed);
		return -1;
	}

	return 0;
}

int audio_play(int rfd, int wfd)
{
	static char	buffer[AUDIO_FILLSZ<<3];
	unsigned int	len;
	int		ratio = 2, i, j, k;
	unsigned short	*buf = (void*)buffer;

	while ((len = read(rfd, buffer, AUDIO_FILLSZ))) {
#if 1		
		for (i = len-1;i >= 0; i--) {
			for (j=i*ratio, k=0; k < ratio; k++,j++) 
				buf[j] = buf[i];
		}
#endif
		//printf("read: %d bytes\n", len);
		if (audio_buffer(wfd, buffer, len*ratio) == -1) {
			perror("write");
			return 3;
		}
	}

	if (audio_flush(wfd) == -1) {
		perror("write");
		return 3;
	}
	return 0;
}

int audio_connect(void)
{
	struct sockaddr_affix	saddr;
	BD_ADDR		bda;
	int		fd, err;

	if (argv[argind] == NULL) {
		printf("parameters missing\n");
		usage();
		return 1;
	}

	err = get_bda(&bda, argv[argind]);
	if (err) {
		printf("Incorrect address given\n");
		return 1;
	}
	printf("Connecting to host %s ...\n", argv[argind]);

	argind++;

	fd = socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCISCO);
	if (fd < 0) {
		printf("Unable to create SCO socket: %d\n", PF_AFFIX);
		return fd;
	}

	saddr.bda = bda;
	saddr.local = BDADDR_ANY;
	err = connect(fd, (struct sockaddr*)&saddr, sizeof(saddr));
	if (err < 0) {
		printf("Unable to connect to remote side\n");
		return err;
	}
	printf("Connected.\n");
	return fd;
}

int cmd_play(struct btctl_command *cmd)
{
	int	rfd, wfd;
	char	buf[256];


	wfd = audio_connect();
	if (wfd < 0)
		return wfd;

	rfd = open(argv[argind], O_RDONLY);
	if (rfd < 0) {
		printf("Unable to open a file %s\n", argv[argind]);
		return rfd;
	}

	printf("Playing...\n");
	for (;;) {
		int		wrote, len;
		char const	*ptr = buf;

		len = read(rfd, buf, 256);
		if (len <= 0)
			break;
		while (len) {
			wrote = write(wfd, ptr, len);
			//printf("wrote: %d\n", wrote);
			if (wrote == -1) {
				if (errno == EINTR)
					continue;
				else
					return -1;
			}
			ptr += wrote;
			len -= wrote;
		}
	}
	pause();
	close(wfd);
	close(rfd);
	return 0;
}

int cmd_listen(struct btctl_command *cmd)
{
	int	err, rfd, wfd;

	if (argv[argind]) {
		rfd = audio_connect();
		if (rfd < 0)
			return rfd;
	} else
		rfd = 0;

	wfd = open(AUDIO_DEVICE, O_WRONLY);
	if (wfd < 0) {
		perror(AUDIO_DEVICE);
		return wfd;
	}

	if (audio_init(wfd) == -1) {
		close(wfd);
		return -1;
	}

	err = audio_play(rfd, wfd);
	if (err) {
		printf("Play error\n");
	}

	close(wfd);
	close(rfd);

	return err;

}


int cmd_record(struct btctl_command *cmd)
{
	int	rfd, count, wfd;
	char	buf[256];

	rfd = audio_connect();
	if (rfd < 0)
		return rfd;

	wfd = open(argv[argind], O_RDWR|O_CREAT);
	if (wfd < 0) {
		printf("Unable to open a file %s\n", argv[argind]);
		return wfd;
	}

	for (;;) {
		count = recv(rfd, buf, 256, 0);
		if (count == 0 || count == -1)
			break;
		//printf("writing %d bytes...\n", count);
		write(wfd, buf, count);
	}

	close(wfd);
	close(rfd);

	return 0;
}


int cmd_addsco(struct btctl_command *cmd)
{
	int	fd;

	fd = audio_connect();
	if (fd < 0)
		return -1;
	pause();
	close(fd);
	return 0;
}


