/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001, 2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
   
   Original Author: Imre Deak <ext-imre.deak@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/* 
   $Id: btsrv-sdp.c,v 1.7 2002/11/11 10:55:04 kds Exp $

   Registering SDP records with the local SDP server.  

   Fixes:
   		Imre Deak <ext-imre.deak@nokia.com>
		Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#include <affix/config.h>

#include <stdlib.h>
#include <string.h>

#include <affix/btcore.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include <affix/sdpsrv.h>

#include "btsrv.h"

static SDPServerHandle	srvHandle = -1;

int sdpreg_init(void)
{
	// initialize the SDP internal data structures

	if (SDPInit(SDP_SVC_PROVIDER) != 0)
		return -1;

	srvHandle = SDPOpenLocalServerConnection();
	if (srvHandle < 0)
		return srvHandle;

	return 0;
}

void sdpreg_cleanup(void)
{
	SDPCloseServerConnection(srvHandle);
	SDPCleanup();
}

int sdpreg_register(service_t *svc)
{
	if (svc->reg_func && svc->reg_func(svc) < 0) {
		BTERROR("Unable to register SDP record for service %s", svc->name);
		return -1;
	}
	return 0;
}

int sdpreg_unregister(service_t *svc)
{
	int		status;
	ServiceRecord	*svcRec = svc->svcRec;

	if (svcRec) {
		status = deleteServiceRecord(srvHandle, svcRec);
		switch (status) {
			case E_OK:
				BTINFO("service %s deleted", svc->name);
				break;
			case E_INVALID_ARGS:
				BTERROR("Error: You cannot delete service %s", svc->name);
				break;
			default:
				BTERROR("Something went wrong. deleteServiceRecord() returned: %d\n", status);
		}
	}

	return 0;
}

ServiceRecord *createPANServiceRecord(UUID16Bit class, slist_t *ptype, uint16_t security, uint16_t type, uint32_t rate)
{
	int		status;
	ServiceRecord	*svcRec;
	SDPData		*attr, *param, *data, *dataSeq;
	slist_t		*pSeq;

	svcRec = createServiceRecord();
	if (svcRec == NULL) {
		BTERROR("createServiceRecord failed");
		return NULL;
	}

	attr = setBrowseGroupList(svcRec);
	if (attr == NULL) {
		BTERROR("setBrowseGroupList failed");
		freeServiceRecord(svcRec);
		return NULL;
	}
	addBrowseGroup(attr, PublicBrowseGroup);

	attr = setProtocolDescriptorList(svcRec);
	if (attr == NULL) {
		BTERROR("setProtocolDescriptorList failed");
		freeServiceRecord(svcRec);
		return NULL;
	}
	
	addProtocolDescriptor(attr, L2CAP_UUID, BNEP_PSM, 2, 0xffff);
	param = addProtocolDescriptor(attr, BNEP_UUID, 0xFFFF, 0, 0x0100);
	/* add supported Network Packet Type */
	/* create root parameter */
	dataSeq = createSDPData(SequenceSize8, NULL);
	/* fill data */
	for (pSeq = ptype; pSeq; pSeq = s_list_next(pSeq)) {
		data = createSDPData(UnsignedInteger_16, pSeq->data);
		addSDPDataSeq(dataSeq, data);
	}
	/* add it to param list */
	addSDPDataSeq(param, dataSeq);	
	
	
	/* set service class list*/
	attr = setServiceClassIDList(svcRec);
	if (attr == NULL) {
		BTERROR("setServiceClassID failed");
		freeServiceRecord(svcRec);
		return NULL;
	}
	addServiceClassID(attr, class);

	/* set profiles list */
	attr = setProfileDescriptorList(svcRec);
	if (attr == NULL) {
		BTERROR("No memory for ProfileDescriptor");
		freeServiceRecord(svcRec);
		return NULL;
	}
	status = addProfileDescriptor(attr, class, 0x0100);
	if (status) {
		freeServiceRecord(svcRec);
		return NULL;
	}
	
	status = setInformationalAttributes(svcRec, "PAN", NULL, NULL);
	if (status != 0) {
		BTERROR("setInformationalAttributes failed");
		freeServiceRecord(svcRec);
		return NULL;
	}

	/* add language attribute */
	attr = setLanguageBaseAttributeIDList(svcRec);
	if (attr == NULL) {
		BTERROR("setLanguageBase failed");
		freeServiceRecord(svcRec);
		return NULL;
	}
	status = addLanguageBase(attr, 0, 0, 0);
	if (status) {
		freeServiceRecord(svcRec);
		return NULL;
	}
		

	addBasicAttribute(svcRec, UnsignedInteger_16, &security, AttrSecurityDescription);

	/* NAP only */
	if (class == NAPServiceClassID) {
		addBasicAttribute(svcRec, UnsignedInteger_16, &type, AttrNetAccessType);
		addBasicAttribute(svcRec, UnsignedInteger_32, &rate, AttrMaxNetAccessRate);
	}

	/* NAP & GN IPv4/6 subnet info - sequence of strings */
	if (class == NAPServiceClassID || class == GNServiceClassID) {

	}

	return svcRec;

}

int sdpreg_serialport(service_t *svc)
{
	int		status = -1;
	ServiceRecord	*svcRec;
	SDPData		*attr;

	/*
	 ** First create a service record handle
	 */
	svcRec = createRFCOMMServiceRecord(svc->srv_ch, 0);
	if (svcRec == NULL) {
		BTERROR("createServiceRecord failed");
		return -1;
	}

	/* set service class list*/
	attr = setServiceClassIDList(svcRec);
	if (attr == NULL) {
		BTERROR("setServiceClassID failed");
		return -1;
	}
	addServiceClassID(attr, SerialPortServiceClassID);

	/* set profiles list */
	attr = setProfileDescriptorList(svcRec);
	if (attr == NULL) {
		BTERROR("No memory for ProfileDescriptor");
		return -1;
	}
	status = addProfileDescriptor(attr, SerialPortServiceClassID, 0x0100);
	if (status) {
		return -1;
	}

	status = setInformationalAttributes(svcRec, svc->name, svc->prov, svc->desc);
	if (status != 0) {
		BTERROR("setInformationalAttributes failed");
		return -1;
	}

	status = registerServiceRecord(srvHandle, svcRec);
	if (status != 0) {
		BTERROR("registerServiceRecord failed");
		return -1;
	}

	svc->svcRec = svcRec;

	return 0;
}

int sdpreg_dialupnetworking(service_t *svc)
{
	int		status = -1;
	ServiceRecord	*svcRec;
	SDPData		*attr;

	/*
	 ** First create a service record handle
	 */
	svcRec = createRFCOMMServiceRecord(svc->srv_ch, 0);
	if (svcRec == NULL) {
		BTERROR("createServiceRecord failed");
		return -1;
	}

	/* set service class list*/
	attr = setServiceClassIDList(svcRec);
	if (attr == NULL) {
		BTERROR("setServiceClassID failed");
		return -1;
	}
	addServiceClassID(attr, DialupNetworkingServiceClassID);
	addServiceClassID(attr, GenericNetworkingServiceClassID);

	/* set profiles list */
	attr = setProfileDescriptorList(svcRec);
	if (attr == NULL) {
		BTERROR("No memory for ProfileDescriptor");
		return -1;
	}
	status = addProfileDescriptor(attr, DialupNetworkingServiceClassID, 0x0100);
	if (status) {
		return -1;
	}

	status = setInformationalAttributes(svcRec, svc->name, svc->prov, svc->desc);
	if (status != 0) {
		BTERROR("setInformationalAttributes failed");
		return -1;
	}

	status = registerServiceRecord(srvHandle, svcRec);
	if (status != 0) {
		BTERROR("registerServiceRecord failed");
		return -1;
	}

	svc->svcRec = svcRec;

	return 0;
}

int sdpreg_lanaccess(service_t *svc)
{
	int		status = -1;
	ServiceRecord	*svcRec;
	SDPData		*attr;

	/* 
	 ** First create a service record handle
	 */
	svcRec = createRFCOMMServiceRecord(svc->srv_ch, 0);
	if (svcRec == NULL)
		return -1;

	/* set service class list*/
	attr = setServiceClassIDList(svcRec);
	if (attr == NULL) {
		BTERROR("setServiceClassID failed");
		return -1;
	}
	addServiceClassID(attr, LANAccessUsingPPPServiceClassID);
	addServiceClassID(attr, GenericNetworkingServiceClassID);

	/* set profiles list */
	attr = setProfileDescriptorList(svcRec);
	if (attr == NULL) {
		BTERROR("No memory for ProfileDescriptor");
		return -1;
	}
	status = addProfileDescriptor(attr, LANAccessUsingPPPServiceClassID, 0x0100);
	if (status) {
		return -1;
	}

	status = setInformationalAttributes(svcRec, svc->name, svc->prov, svc->desc);
	if (status != 0)
		return -1;

	status = registerServiceRecord(srvHandle, svcRec);
	if (status != 0)
		return -1;

	svc->svcRec = svcRec;

	return 0;
}


int sdpreg_obex_filetransfer(service_t *svc)
{
	int		status = -1;
	ServiceRecord	*svcRec;
	SDPData		*attr;

	/*
	 ** First create a service record handle
	 */
	svcRec = createRFCOMMServiceRecord(svc->srv_ch, 1);
	if (svcRec == NULL)
		return -1;

	/* set service class list*/
	attr = setServiceClassIDList(svcRec);
	if (attr == NULL) {
		BTERROR("setServiceClassID failed");
		return -1;
	}
	addServiceClassID(attr, ObexFileTransferServiceClassID);

	/* set profiles list */
	attr = setProfileDescriptorList(svcRec);
	if (attr == NULL) {
		BTERROR("No memory for ProfileDescriptor");
		return -1;
	}
	status = addProfileDescriptor(attr, ObexFileTransferServiceClassID, 0x0100);
	if (status) {
		return -1;
	}

	status = setInformationalAttributes(svcRec, svc->name, svc->prov, svc->desc);
	if (status != 0)
		return -1;

	status = registerServiceRecord(srvHandle, svcRec);
	if (status != 0)
		return -1;

	svc->svcRec = svcRec;

	return 0;
}

int sdpreg_obex_objectpush(service_t *svc)
{
	int		status = -1;
	ServiceRecord	*svcRec;
	SDPData		*attr;

	/*
	 ** First create a service record handle
	 */
	svcRec = createRFCOMMServiceRecord(svc->srv_ch, 1);
	if (svcRec == NULL)
		return -1;

	/* set service class list*/
	attr = setServiceClassIDList(svcRec);
	if (attr == NULL) {
		BTERROR("setServiceClassID failed");
		return -1;
	}
	addServiceClassID(attr, ObexObjectPushServiceClassID);

	/* set profiles list */
	attr = setProfileDescriptorList(svcRec);
	if (attr == NULL) {
		BTERROR("No memory for ProfileDescriptor");
		return -1;
	}
	status = addProfileDescriptor(attr, ObexObjectPushServiceClassID, 0x0100);
	if (status) {
		return -1;
	}

	status = setInformationalAttributes(svcRec, svc->name, svc->prov, svc->desc);
	if (status != 0)
		return -1;

	status = registerServiceRecord(srvHandle, svcRec);
	if (status != 0)
		return -1;

	svc->svcRec = svcRec;

	return 0;
}


