/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001,2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Original Author: Imre Deak <ext-imre.deak@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btsrv.c,v 1.41 2002/11/14 10:04:03 kds Exp $

   User space daemon for making local services accessable to remote devices   

   Fixes:
   		Imre Deak <ext-imre.deak@nokia.com>
		Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#include <affix/config.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/errno.h>

#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <getopt.h>
#include <string.h>
#include <sys/mman.h>

/* bluetooth stuff */
#include <affix/bluetooth.h>
#include <affix/btcore.h>

#include "btsrv.h"

#define DEF_CONFIG_FILE		"/etc/affix/btsrv.conf"

char	*config_file = NULL;
int	argind;
int	usesdp = 1;
int	usegui = 1;
int	usesvc = 1;	/* start services */
int	usekeys = 0;

/* global configuration block */
service_t	services[MAX_SERVICE_NUM];
device_t	devices[MAX_DEVICE_NUM];
btdev_list	btdevs;

int 		services_num;
int		devices_num;
int		efd, mfd;
char		confdir[80];

int btsrv_devconf(int fd);


void printusage(void)
{
	printf("usage: btsrv "
			"[-d] "
			"[-v] "
			"[--nogui | -g] "
			"[--nosdp | -s] "
			"--nosvc "
			"[--config config_file | -C config_file] "
			"[--manage-keys]"
			"\n"
	      );
}

int start_service(service_t *svc)
{
	struct sockaddr_affix	saddr;
	int		err, fd;
	socklen_t	len = sizeof(saddr);

	fd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_RFCOMM);
	if (fd < 0) {
		BTERROR("Unable to create RFCOMM socket");
		return -1;
	}

	if (fcntl(fd, F_SETFL, O_NONBLOCK) != 0) {
		BTERROR("Unable to set non blocking mode for RFCOMM socket");
		return -1;
	}

	if (svc->flags & SRV_FLAG_RFCOMM_TTY) {
		err = rfcomm_set_type(fd, RFCOMM_BTY);
		if (err < 0) {
			BTERROR("Unable to set RFCOMM interface type to tty");
			return -1;
		}
	}

	saddr.family = AF_AFFIX;
	saddr.port = 0;		/* dynamically allocated */
	saddr.bda = BDADDR_ANY;
	err = bind(fd, (struct sockaddr*)&saddr, sizeof(saddr));
	if (err < 0) {
		BTERROR("Unable to bind address");
		return -1;
	}

	setsockopt(fd, SOL_AFFIX, BTSO_SECURITY, &svc->security, sizeof(int));
	//printf("sec_level: %d\n", svc->security);

	err = getsockname(fd, (struct sockaddr*)&saddr, &len);
	if (err < 0) {
		BTERROR("Unable to get address info");
		return -1;
	}

	BTINFO("Bound service %s to server channel %d", svc->name, saddr.port);

	err = listen(fd, 5);
	if (err < 0) {
		BTERROR("Unable to listen for connection requests");
		return -1;
	}

	svc->srv_fd = fd;
	svc->srv_ch = saddr.port;

#if defined(CONFIG_AFFIX_SDP)
	if (usesdp && sdpreg_register(svc) < 0 )
		return -1;
#endif	
	svc->running = 1;
	return 0;
}

void stop_service(service_t *svc)
{
	if (!svc->running)
		return;
#if defined(CONFIG_AFFIX_SDP)
	if (usesdp)
		sdpreg_unregister(svc);
#endif
	close(svc->srv_fd);
}

int start_all_services(void)
{
	int	i, j;

	for (i = 0; i < services_num; i++)
		if (start_service(&services[i]) < 0)
			goto fail;
	return 0;
fail:
	for (j = 0; j < i; j++)
		stop_service(&services[i]);
	return -1;
}

void stop_all_services(void)
{
	int	i;

	for (i = 0; i < services_num; i++)
		stop_service(&services[i]);
}

char *btsrv_format_cmd(const char *cmd, int conid, int line, BD_ADDR *bda, int srv_ch)
{
        char    *formatted_cmd;
	int	i, j, res;
 
        formatted_cmd = (char *)malloc(MAX_CMD_LEN + 1);
        if (formatted_cmd == NULL)
                return NULL;
        i = j = 0;
        while (cmd[i]) {
                switch (cmd[i]) {
                case '%':
                        switch(cmd[i + 1])
                        {
                        case 'a':
                                res = snprintf(&formatted_cmd[j], MAX_CMD_LEN - j, "%s", bda2str(bda));
                                break;
                        case 'l':
                                res = snprintf(&formatted_cmd[j],
					      MAX_CMD_LEN - j, "%d", line);
                                break;
			case 'i':
                                res = snprintf(&formatted_cmd[j],
					      MAX_CMD_LEN - j, "%d", conid);
                                break;

                        case 'c':
                                res = snprintf(&formatted_cmd[j],
					       MAX_CMD_LEN - j, "%d", srv_ch);
                                break;
                        case '%':
                                formatted_cmd[j] = '%';
				res = j < MAX_CMD_LEN ? 1 : -1;
                                break;
			default:
				res = -1;
                        }
                        i++;	/* skip % sign */
                        break;
                default:
                        formatted_cmd[j] = cmd[i];
			res = j < MAX_CMD_LEN ? 1 : -1;
                }
		if (res == -1) {
			free(formatted_cmd);
			return NULL;
		}
		j += res;
                i++;
        }
        formatted_cmd[j] = '\0';
        return formatted_cmd;
}


int execute_cmd(char *cmd, int sock_fd, int srv_ch, BD_ADDR *bd_addr, int flags)
{
	char *formatted_cmd;
	int  line;
	int  fd = -1;
	int  conid;

	conid = sock_fd;
	if (flags & SRV_FLAG_RFCOMM_TTY) {
		char dev[10];

		line = rfcomm_open_tty(sock_fd, -1);
		close(sock_fd);
		if (line < 0) {
			BTERROR("Unable to bind a socket to virtual port device");
			return -1;
		}
		sprintf(dev, "/dev/bty%d", line);
		BTINFO("Socket bound to virtual port device %s", dev);
		if (flags & SRV_FLAG_STD) {
			fd = open(dev, O_RDWR);
			if (fd < 0) {
				BTERROR("Unable to open port device %s", dev);
				return -1;
			}
		}
	} else {
		fd = sock_fd;
		line = -1;
	}
	if (flags & SRV_FLAG_STD) {
		BTINFO("Socket multiplexed to stdin/stdout");
		if (fd != 0){
			close(0);
			if (dup2(fd, 0) < 0) {
				BTERROR("Unable to duplicate socket descriptor to stdin\n");
				return -1;
			}
		}
		if (fd != 1) {
			close(1);
			if (dup2(fd, 1) < 0) {
				BTERROR("Unable to duplicate socket descriptor to stdout\n");
				return -1;
			}
		}
		if (fd != 0 && fd != 1)
			close(fd);
		close(2);
		if (open("/dev/null", O_RDWR) != 2) {
			BTERROR("Cannot open /dev/null to stderr");
			return -1;
		}
	}
	formatted_cmd = btsrv_format_cmd(cmd, conid, line, bd_addr, srv_ch);
	if (formatted_cmd) {
		BTINFO("Execute %s", formatted_cmd);
		/*  the following does not return normally  */
		execl("/bin/sh", "sh", "-c", formatted_cmd, NULL);
		BTERROR("Exec error");
		return -1;
	}
	BTERROR("Parse error in the command line");
	return -1;
}
/*
 * Event processing block
 */

int event_pin_code_request(struct PIN_Code_Request_Event *evt, int hd)
{
	int		fd, err;
	FILE		*fp;
	char		pin[32], cmdline[100];
	char		name[248];

	if (!usegui)
		return 0;

	fd = hci_open_id(hd);
	if (fd < 0)
		return fd;
#if 0
	{
		/* get device name first */
		INQUIRY_ITEM	dev;
		dev.bda = evt->bda;
		dev.PS_Repetition_Mode = 0x00;
		dev.PS_Mode = 0x00;
		dev.Clock_Offset = 0x00;
		err = HCI_RemoteNameRequest(fd, &dev, name);
		if (err) {
			close(fd);
			BTERROR("Name request failed");
			return err;
		}
	}
#else
	name[0] = '\0';
#endif

	sprintf(cmdline, "/etc/affix/btsrv-gui pin %s \"%s\"", bda2str(&evt->bda), name);
	BTDEBUG("cmdline: [%s]", cmdline);
	fp = popen(cmdline, "r");
	if (!fp) {
		BTERROR("popen() failed");
		close(fd);
		return -1;
	}
	err = fscanf(fp, "%s", pin);
	if (err == EOF) {
		BTERROR("pipe read failed");
		pclose(fp);
		err = HCI_PINCodeRequestNegativeReply(fd, &evt->bda);
		close(fd);
		return 0;
	}
	BTDEBUG("Got PIN code from pipe: %s, len: %d", pin, strlen(pin));
	pclose(fp);

	err = HCI_PINCodeRequestReply(fd, &evt->bda, strlen(pin), pin);
	if (err) {
		BTERROR("unable to set pin code: %d", err);
	}

	close(fd);

	return 0;
}

int event_link_key_request(struct Link_Key_Request_Event *evt, int hd)
{
	int		fd, err, ok = 0;
	btdev_struct	*btdev;

	fd = hci_open_id(hd);
	if (fd < 0)
		return fd;

	btdev_cache_reload(&btdevs);
	btdev = btdev_cache_lookup(&btdevs, &evt->bda);
	if (btdev && (btdev->flags & BTDEV_KEY))
		ok = 1;

	if (ok) {
		err = HCI_LinkKeyRequestReply(fd, &evt->bda, btdev->link_key);
	} else {
		err = HCI_LinkKeyRequestNegativeReply(fd, &evt->bda);
	}
	if (err) {
		BTERROR("unable to set link key: %d", err);
	}
	btdev_cache_unlock(&btdevs);

	hci_close(fd);

	return err;
}

int event_link_key_notification(struct Link_Key_Notification_Event *evt, int hd)
{
	btdev_struct	*btdev;

	if (!usekeys)
		return 0;
	btdev_cache_reload(&btdevs);
	btdev = btdev_cache_add(&btdevs, &evt->bda);
	if (btdev == NULL)
		return -1;
	btdev->key_type = evt->Key_Type;
	memcpy(btdev->link_key, evt->Link_Key, 16);
	btdev->flags |= BTDEV_KEY;
	btdev_cache_save(&btdevs);
	
	return 0;
}


int event_handler(int fd)
{
	__u8	event[EVENT_MAX_SIZE];
	HCI_Event_Packet_Header	*hdr = (void*)event;
	int	hd;

	hci_recv_event(fd, event, &hd);
	BTDEBUG("got event: %#x", hdr->EventCode);
	switch (hdr->EventCode) {
		case HCI_E_LINK_KEY_REQUEST:
			event_link_key_request((void*)event, hd);
			break;
		case HCI_E_LINK_KEY_NOTIFICATION:
			event_link_key_notification((void*)event, hd);
			break;
		case HCI_E_PIN_CODE_REQUEST:
			event_pin_code_request((void*)event, hd);
			break;

		default:
	}

	return 0;
}

int message_handler(int fd)
{
	__u8			buf[MSG_MAX_SIZE];
	struct hci_msg_hdr	*msg = (void*)buf;
	int			err;

	hci_recv_msg(fd, buf);
	BTDEBUG("got message: %x", msg->opcode);
	if (msg->opcode == HCICTL_STATE_CHANGE) {
		struct hci_state_change	*cmd = (void*)msg;

		if (cmd->event == HCIDEV_UP) {
			BTDEBUG("device is up\n");
			err = btsrv_devconf(cmd->hd);
			if (err) {
				BTERROR("Unable to configure device: %d", cmd->hd);
				return err;
			}
		}
	}
	return 0;
}

int handle_input(void)
{
	int	i, err = 0;
	int	fd_cnt;
	fd_set	fds_read;
	int	max_fd;

	for(;;) {
		FD_ZERO(&fds_read);
		max_fd = -1;
		/* set service fd */
		for (i = 0; i < services_num; i++) {
			if (services[i].running) {
				FD_SET(services[i].srv_fd, &fds_read);
				if (max_fd < services[i].srv_fd)
					max_fd = services[i].srv_fd;
			}
		}
		/* set event fd */
		FD_SET(efd, &fds_read);
		if (max_fd < efd)
			max_fd = efd;
		FD_SET(mfd, &fds_read);
		if (max_fd < mfd)
			max_fd = mfd;
		/* poll data */
		fd_cnt = select(max_fd + 1, &fds_read, NULL, NULL, NULL);
		if (fd_cnt <= 0) {
			BTERROR("Select error");
			err = -1;
			goto exit;
		}
		if (FD_ISSET(efd, &fds_read)) {
			event_handler(efd);
		}
		if (FD_ISSET(mfd, &fds_read)) {
			message_handler(mfd);
		}
		for (i = 0; fd_cnt && (i < services_num); i++) {
			if (services[i].running && FD_ISSET(services[i].srv_fd, &fds_read)) {
				int 			newsock;
				struct sockaddr_affix	caddr;
				socklen_t		calen = sizeof(caddr);
				int			pid;

				newsock = accept(services[i].srv_fd, (struct sockaddr*)&caddr, &calen);
				if (newsock == -1) {
					BTERROR("Accept error");
					continue;
				}
				BTINFO("Connection from %s\nserver channel %d (%s Profile)",
						bda2str(&caddr.bda), caddr.port, services[i].name);

				if (services[i].cmd[0] != '\0') {
					pid = fork();
					if (pid < 0) {
						BTERROR("Fork error");
						continue;
					}
					if (pid == 0) {
						int res, j;

						for (j = 0; j < services_num; j++)
							close(services[j].srv_fd);
						res = execute_cmd(services[i].cmd, newsock,
								caddr.port, &caddr.bda, 
								services[i].flags);
						/*  In case of error.. */
						exit(res);
					}
				}
				close(newsock);
				fd_cnt--;
			}
		}
	}
exit:
	return err;
}

/*
 * read HCI events and kernel messages
 */

int start_event_handler(void)
{
	__u64		mask = 0;
#if 0
	btdev_struct	*entry;
	int		i;
#endif

	efd = hci_open_event();
	if (efd < 0) {
		BTERROR("Unable to open hci(NULL)");
		return efd;		
	}
	mfd = hci_open_mgr();
	if (mfd < 0) {
		BTERROR("Unable to open mgr");
		close(efd);
		return mfd;
	}
	if (usegui) {
		hci_set_mode(mfd, AFFIX_MODE_PIN);
		mask |= PIN_CODE_REQUEST_MASK;
	}
	if (usekeys) {
		hci_set_mode(mfd, AFFIX_MODE_KEY);
		mask |= LINK_KEY_NOTIFICATION_MASK | LINK_KEY_REQUEST_MASK;
#if 0
		/* add all keys */
		for (i = 0; (entry = s_list_nth_data(btdevs.head, i)); i++) {
			if (entry->flags & BTDEV_KEY)
				hci_add_key(&entry->bda, entry->key_type, entry->link_key);
		}
#endif
	}

	if (mask)
		hci_listen_event(efd, mask);

	return 0;
}

void stop_event_handler(void)
{
	close(mfd);
	close(efd);
}

int btsrv_setdev(int fd, device_t *dev)
{
	int	err;

	err = hci_set_secmode(fd, dev->security);
	if (err < 0) {
		BTERROR("hci_set_secmode() failed\n");
		return err;
	}
	err = hci_set_role(fd, dev->role);
	if (err < 0) {
		BTERROR("hci_set_role() failed\n");
		return err;
	}
	err = hci_set_pkttype(fd, dev->pkt_type);
	if (err < 0) {
		BTERROR("hci_set_pkttype() failed\n");
		return err;
	}
	err = HCI_WriteClassOfDevice(fd, dev->cod);
	if (err) {
		BTERROR("WriteScanEnable error\n");
		return err;
	}
	if (dev->btname) {
		err = HCI_ChangeLocalName(fd, dev->btname);
		if (err) {
			BTERROR("WriteScanEnable error\n");
			return err;
		}
	}
	err = HCI_WriteScanEnable(fd, dev->scan);
	if (err) {
		BTERROR("WriteScanEnable error\n");
		return err;
	}
	return 0;
}

int btsrv_devconf(int hd)
{
	int			err = 0, i, fd;
	device_t		*dev;
	struct hci_dev_attr	da;

	err = hci_get_attr_id(hd, &da);
	if (err)
		return err;
	fd = _hci_open(da.name);
	if (fd < 0) {
		BTERROR("_hci_open_id() failed: %s\n", da.name);
		return fd;
	}
	for (i = 0; i < devices_num; i++) {
		dev = &devices[i];
		if ((i == 0 && dev->name[0] == '*') ||
				(strcasecmp(dev->name, da.name) == 0)) {
			err = btsrv_setdev(fd, dev);
			if (err)
				break;
		}
	}
	close(fd);
	return err;
}

			
int btsrv_devinit(void)
{
	int	i, err, num;
	int	devs[HCI_MAX_DEVS];
	struct hci_dev_attr	da;

	num = hci_get_devs(devs);
	if (num < 0) {
		BTERROR("Unable to get device info\n");				
		return num;
	}
	for (i = 0; i < num; i++) {
		err = hci_get_attr_id(devs[i], &da);
		if (err < 0) {
			BTERROR("hci_get_attr() failed: %d\n", devs[i]);
			continue;
		}
		if (!(da.flags & HCI_FLAGS_UP))
			continue;
		err = btsrv_devconf(da.hd);
		if (err)
			return err;
	}
	return 0;
}

int btsrv_cleanup(void)
{
	int		i;
	device_t	*dev;

	stop_event_handler();
	stop_all_services();
	BTINFO("Terminating ..");
#if defined(CONFIG_AFFIX_SDP)
	sdpreg_cleanup();
#endif
	free(config_file);
	for (i = 0; i < services_num; i++) {
		if (services[i].name)
			free(services[i].name);
		if (services[i].prov)
			free(services[i].prov);
		if (services[i].desc)
			free(services[i].prov);
	}

	for (i = 0; i < devices_num; i++) {
		dev = &devices[i];
		if (dev->btname)
			free(dev->btname);
	}
	return 0;
}

void signal_handler(int sig)
{
	BTINFO("Sig handler : %d", sig);
	btsrv_cleanup();
	exit(0);
}

int main(int argc, char **argv)
{
	int	c, lind, err, btdaemon;
	char	cachefile[80];

	struct option	opts[] = {
		{"help", 0, 0, 'h'},
		{"daemon", 0, 0, 'd'},
		{"nogui", 0, 0, 'g'},
		{"config", 1, 0, 'C'},
		{"nosdp", 0, 0, 's'},
		{"debug", 0, 0, 'v'},
		{"nosvc", 0, 0, 'r'},
		{"manage-keys", 0, 0, 'k'},
		{0, 0, 0, 0}
	};

	openlog(NULL/*argv[0]*/, LOG_PERROR, LOG_DAEMON);

	BTINFO("%s %s started!!!", argv[0], affix_version);
	
	btdaemon = 0;

	btlog_mask = BTDEBUG_MODULE;
	for (;;) {
		c = getopt_long(argc, argv, "hdgvC:", opts, &lind);
		if (c == -1)
			break;

		switch (c) {
			case 'h':
				printusage();
				return 0;
				break;
			case 'd':	
				btdaemon = 1;
				break;
			case 'g':
				usegui = 0;
				break;
			case 'C':
				config_file = strdup(optarg);
				break;
			case 's':
				usesdp = 0;
				break;
			case 'v':
				btlog_mask = BTDEBUG_MODULE;
				break;
			case 'r':
				usesvc = 0;
				break;
			case 'k':
				usekeys = 1;
				break;
			case '?':
				fprintf(stderr, "Unknown option: %c\n", optopt);
				return -1;
				break;
		}
	}

	argind = optind;

	if (getuid() == 0) {
		/* root -> put the stuff to /var/spool/affix */
		sprintf(confdir, "/var/spool/affix");
	} else
		sprintf(confdir, "%s/.bluetooth", getenv("HOME"));
	err = rmkdir(confdir, 0700);
	if (err < 0) {
		BTERROR("Unable to create dir [%s]: %s", confdir, strerror(errno));
		return 1;
	}
	sprintf(cachefile, "%s/%s", confdir, "devcache");
	btdev_cache_load(cachefile, &btdevs);

	/* test gui access */
	if (usegui) {
		err = system("/etc/affix/btsrv-gui try 2>/dev/null");
		if (err) {
			printf("GUI is not available... PIN helper disabled\n");
			usegui = 0;
		}
	}

#if defined(CONFIG_AFFIX_SDP)
	if (usesdp) {
		BTINFO("SDP server is used\n");
		if (sdpreg_init() != 0) {
			BTERROR("SDP initialization failed");
			return -1;
		}
	}
#else
	usesdp = 0;
#endif

	/* initialize btsrv stuff */
	if (config_file == NULL)
		config_file = strdup(DEF_CONFIG_FILE);

	if (btsrv_read_config(config_file, &services_num, &devices_num)) {
		BTERROR("Invalid config file: %s", config_file);
		return -1;
	}

	if (usesvc && services_num < 1) {
		BTERROR("No services found in %s", config_file);
		return -1;
	}

	/* initialize devices */
	err = btsrv_devinit();
	if (err) {
		BTERROR("Unable to initialize devices\n");
		return -1;
	}

	signal(SIGCHLD, SIG_IGN);
	if (start_event_handler() < 0) {
		BTERROR("Cannot start event handler");
		return -1;
	}

	if (usesvc && start_all_services() < 0) {
		BTERROR("Cannot start services");
		return -1;
	}

	/*    Exit the parent process if run as a daemon */
	if (btdaemon && become_daemon(1)) {
		return 0;
	}

	/* install signal handlers to unregister services from SDP */
	signal(SIGINT, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGABRT, signal_handler);
	signal(SIGQUIT, signal_handler);

	if (handle_input() < 0) {
		BTERROR("Unable to listen to incoming connections");
	}
	btsrv_cleanup();

	return 0;
}

