/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2002 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: bthfp.c,v 1.14 2002/11/27 18:44:36 kds Exp $

   Hands Free profile client library

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

*/

#include <affix/config.h>

#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>

#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <affix/bluetooth.h>
#include <affix/btcore.h>

// SDP stuff
#include <affix/sdp.h>
#include <affix/sdpsrv.h>
#include <affix/sdpclt.h>

// OBEX stuff
//#include <obex.h>
//#include "obex_server.h"
//#include "obex_client.h"

// Local include
#include <affix/bthfp.h>

static SDPServerHandle	srvHandle = -1;


/* internal declarations */
int hfp_service_register(hfp_t *handle);
int hfp_service_unregister(hfp_t *handle);


hfp_t *hfp_init(hfp_event_t eventcb, int flags)
{
	hfp_t		*handle = NULL;

	handle = malloc(sizeof(hfp_t));
	if (handle == NULL)
		return NULL;

	handle->eventcb = eventcb;

	return handle;
}

void hfp_cleanup(hfp_t *handle)
{
	hfp_service_unregister(handle);
	free(handle);
}

int getSupportedFeatures(ServiceRecord *svcRec, uint16_t *SupportedFeatures)
{
	GetBasicAttribute(AttrSupportedFeatures, SupportedFeatures, uint16);
}

int hfp_service_search(hfp_t *handle)
{
	int		status = -1, i, s, p;
	uint16_t	count = 0;
	slist_t		*pSeq = NULL;
	slist_t		*attrList = NULL;
	uint32_t	attr1 = AttrServiceRecordHandle;
	uint32_t	attr2 = AttrProtocolDescriptorList;
	uint32_t	attr3 = AttrSupportedFeatures;
	slist_t		*searchList = NULL;
	UUID		uuid[2];
	AccessProtocols	*pAccessProtocols = NULL;
	SDPServerHandle srvHandle;
	ServiceRecord	*svcRec;
	struct sockaddr_affix	sa;

	status = SDPInit(0);
	if (status != E_OK)
		return status;

	sa.family = PF_AFFIX;
	sa.bda = handle->peer.bda;
	sa.local = BDADDR_ANY;
	srvHandle = SDPOpenServerConnection(&sa);
	if (srvHandle < 0)
		return srvHandle;

	s_list_append(&attrList, &attr1);
	s_list_append(&attrList, &attr2);
	s_list_append(&attrList, &attr3);

	makeUUIDFrom16Bits(&uuid[0], AGHandsFreeServiceClassID);
	makeUUIDFrom16Bits(&uuid[1], SerialPortServiceClassID);

	for (s = 0; s < 2; s++ ) {
		s_list_append(&searchList, &uuid[s]);
		status = SDPServiceSearchRequest(srvHandle, searchList, 1, &pSeq, &count);
		BTDEBUG("Status: %d, count: %d, len: %d", status, count, (int)s_list_length(pSeq));
		s_list_free(&searchList);
		if ((status == 0) && (count > 0))
			break;
	}
	if ((status == 0) && (count > 0)) {
		for (i = 0; i < (int)s_list_length(pSeq); i++) {
			ServiceRecordHandle svcRecHandle = *(ServiceRecordHandle*)s_list_nth_data(pSeq, i);
			BTDEBUG("Geting attributes, handle: %#x", svcRecHandle);
			status = SDPServiceAttributeRequest(srvHandle, svcRecHandle,
					IndividualAttributes, attrList, 0xFFFF, &svcRec, &count);
			if (status != 0)
				goto exit;

			if (getAccessProtocols(svcRec, &pAccessProtocols) == E_OK) {
				for (s = 0; s < (int)s_list_length(pAccessProtocols->pSequence); s++ ) {
					slist_t *protDescSeq = (slist_t*)s_list_nth_data(pAccessProtocols->pSequence, s);
					for (p = 0; p < (int)s_list_length(protDescSeq); p++ ) {
						ProtocolDescriptor *pPD = (ProtocolDescriptor*)s_list_nth_data(protDescSeq, p);
						if (UUID2value(&pPD->protocolId) == RFCOMM_UUID) {
							if (pPD->portNumber) {
								handle->peer.port = pPD->portNumber;
							}
							goto next;
						}
					}
				}
			}
next:
			if (getSupportedFeatures(svcRec, &handle->SupportedFeatures)) {
				BTDEBUG("Features found: %hx", handle->peerSupportedFeatures);
			}

		}
	}
exit:
	SDPCloseServerConnection(srvHandle);
	
	return status?status:handle->peer.port;

}

/*
 * Register HandsFree service
 */
int hfp_service_register(hfp_t *handle)
{
	int		status = -1;
	ServiceRecord	*svcRec;
	SDPData		*attr;
	
	/*
	 ** First create a service record handle
	 */
	svcRec = createRFCOMMServiceRecord(handle->local.port, 0);
	if (svcRec == NULL) {
		BTERROR("createServiceRecord failed");
		return -1;
	}

	/* set service class list*/
	attr = setServiceClassIDList(svcRec);
	if (attr == NULL) {
		BTERROR("setServiceClassID failed");
		return -1;
	}
	addServiceClassID(attr, HandsFreeServiceClassID);
#if 1
	addServiceClassID(attr, SerialPortServiceClassID);
	addServiceClassID(attr, HeadsetServiceClassID);
//	addServiceClassID(attr, GenericAudioServiceClassID);
#endif

	/* set profiles list */
	attr = setProfileDescriptorList(svcRec);
	if (attr == NULL) {
		BTERROR("No memory for ProfileDescriptor");
		return -1;
	}
	status = addProfileDescriptor(attr, HandsFreeServiceClassID, 0x0100);
	if (status) {
		return -1;
	}

	status = setInformationalAttributes(svcRec, handle->name, NULL, NULL);
	if (status != 0) {
		BTERROR("setInformationalAttributes failed");
		return -1;
	}

	/* set Supported Features */
	addBasicAttribute(svcRec, UnsignedInteger_16, &handle->SupportedFeatures, 
			AttrSupportedFeatures);

	status = registerServiceRecord(srvHandle, svcRec);
	if (status != 0) {
		BTERROR("registerServiceRecord failed");
		return -1;
	}

	handle->svcRec = svcRec;

	return 0;
}


int hfp_service_unregister(hfp_t *handle)
{
	int	status;
	
	if (handle->svcRec) {
		status = deleteServiceRecord(srvHandle, handle->svcRec);
		switch (status) {
			case E_OK:
				BTINFO("service %s deleted", handle->name);
				break;
			case E_INVALID_ARGS:
				BTERROR("Error: You cannot delete service %s", handle->name);
				break;
			default:
				BTERROR("Something went wrong. deleteServiceRecord() returned: %d", status);
		}
		handle->svcRec = NULL;
	}
 
	return 0;
}


/*
 * Create 
 */
int hfp_server_register(hfp_t *handle)
{
	int			err, fd;
	struct	sockaddr_affix	sa;
	socklen_t		len = sizeof(sa);

	fd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_RFCOMM);
	if (fd < 0) {
		BTERROR("Unable to create RFCOMM socket");
		return -1;
	}
	
	if (fcntl(fd, F_SETFL, O_NONBLOCK) != 0) {
		BTERROR("Unable to set non blocking mode for RFCOMM socket");
		return -1;
	}

	sa.family = AF_AFFIX;
	sa.port = 0;		/* dynamically allocated */
	sa.bda = BDADDR_ANY;
	err = bind(fd, (struct sockaddr*)&sa, sizeof(sa));
	if (err < 0) {
		BTERROR("Unable to bind address");
		return -1;
	}
	
	err = getsockname(fd, (struct sockaddr*)&sa, &len);
	if (err < 0) {
		BTERROR("Unable to get address info");
		return -1;
	}

	BTINFO("Bound service \"%s\" to server channel %d", handle->name, sa.port);

	err = listen(fd, 5);
	if (err < 0) {
		BTERROR("Unable to listen for connection requests");
		return -1;
	}

	handle->sfd = fd;
	handle->local = sa;

#if defined(CONFIG_AFFIX_SDP)
	err = SDPInit(SDP_SVC_PROVIDER);
	if (err) {
		BTERROR("Unable to initialize SDP");
		return -1;
	}
	sa.family = PF_AFFIX;
	sa.bda = handle->peer.bda;
	sa.local = BDADDR_ANY;
	srvHandle = SDPOpenServerConnection(&sa);
	if (srvHandle < 0)
		return srvHandle;

	if (sdpmode && hfp_service_register(handle) < 0 )
		return -1;
#endif	
	return 0;

	
}

int hfp_connect(hfp_t *handle, void *addr)
{
	int		err, fd;
	BD_ADDR		bda;
	int		sch;
	struct sockaddr_affix	*sa = addr;

	bda = sa->bda;
	handle->peer = *sa;

	BTDEBUG("Connecting to host %s ...", bda2str(&bda));

#if defined(CONFIG_AFFIX_SDP)
	if (sa->port == 0) {
		sch = hfp_service_search(handle);
		if (sch > 0)
			BTDEBUG("Service found on channel %d", sch);
		else if ( sch == 0 ) {
			BTDEBUG("Service is not available");
			return -1;
		} else {
			BTDEBUG("Unable to get service channel: %d", sch);
			return -1;
		}
	} else
#endif	
		sch = sa->port;

	sleep(1);
	fd = socket(PF_AFFIX, SOCK_STREAM, BTPROTO_RFCOMM);
	if (fd < 0) {
		BTDEBUG("Unable to create RFCOMM socket: %d", PF_AFFIX);
		return -1;
	}

	handle->peer.port = sch;
	handle->peer.local = BDADDR_ANY;
	err = connect(fd, (struct sockaddr*)&handle->peer, sizeof(*sa));
	if (err < 0) {
		BTDEBUG("Unable to connect to remote side");
		close(fd);
		return -1;
	}
	handle->fd = fd;
	
	return 0;
}

int hfp_handle_input(hfp_t *handle)
{
	int	status, timeout = -1;
	
	struct pollfd	fds[2];

	fds[SERVERFD].fd = handle->sfd;
	fds[SERVERFD].events = POLLIN;

	fds[CLIENTFD].fd = handle->fd;
	fds[CLIENTFD].events = POLLIN;
	

	status = poll(fds, 2, timeout);
	if (status > 0) {
		if ((fds[CLIENTFD].revents | fds[SERVERFD].revents) & (POLLERR | POLLHUP | POLLNVAL))
			;
		else {
			if (fds[SERVERFD].revents) {
				if (fds[SERVERFD].revents & POLLIN) { /* incomming connection */
				}
			}
			if (fds[CLIENTFD].revents) {
				if (fds[CLIENTFD].revents & POLLIN) {/* data from peer */
				}
			}
		}
	} else if (status < 0)
		return status;

	return 0;
}

