/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001,2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Original Author: Guruprasad Krishnamurthy <kgprasad@hotmail.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: cstate.c,v 1.14 2002/11/04 08:04:06 kds Exp $

   Fixes:
   		Dmitry Kasatkin		: cleanup, fixes
*/

#include <affix/config.h>

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdint.h>

#include "cstate.h"
#include "utils.h"

static slist_t	*cacheList;

void printContinuationState(ContinuationState *pCState)
{
	BTDEBUG("[%s-%d] : ", __FUNCTION__, __LINE__);
	BTDEBUG("CState TS : 0x%lx", pCState->timestamp);
	BTDEBUG("CState max bytes : %d", pCState->cStateValue.maxBytesSent);
	BTDEBUG("CState last index : %d", pCState->cStateValue.lastIndexSent);
}

void cacheListPrint(void * value, void * data)
{
	BTDEBUG("Value : %p", (char *)value);
}


void initContinuationStateCache(void)
{
	cacheList = NULL;
}

ContinuationState *getContinuationState(char *buffer)
{
	char		*pdata = buffer;
	uint8_t 	cStateSize = 0;
	ContinuationState *pCState = NULL;

	/*
	 ** Check if continuation state exists, if yes attempt
	 ** to get response remainder from cache, else send error
	 */
	cStateSize = __get_u8(pdata);
	BTDEBUG("Continuation State size : %d", cStateSize);
	pdata += sizeof(uint8_t);
	if (cStateSize != 0) {
		pCState = (ContinuationState *)pdata;
		BTDEBUG("Cstate TS : 0x%lx", pCState->timestamp);
		BTDEBUG("Bytes sent : %d", pCState->cStateValue.maxBytesSent);
	}                 

	return pCState;
}

int setContinuationState(char *pdata, ContinuationState *pCState)
{
	int length = 0;

	if (pCState == NULL) {
		__put_u8(pdata, 0);
		pdata += sizeof(uint8_t);
		length += sizeof(uint8_t);
	} else {
		BTDEBUG("Non null CState id : 0x%lx", pCState->timestamp);
		__put_u8(pdata, sizeof(ContinuationState));
		pdata += sizeof(uint8_t);
		length += sizeof(uint8_t);
		memcpy(pdata, pCState, sizeof(ContinuationState));
		length += sizeof(ContinuationState);
	}
	return length;
}

int setContinuationStatePDU(PDUForm *pduForm, ContinuationState *pCState)
{
	int length = 0;
	char *pdata = NULL;

	pdata = pduForm->data + pduForm->dataSize;
	length = setContinuationState(pdata, pCState);
	pduForm->dataSize += length;
	
	return length;
}

long createContinuationStateBuffer(int fd, PDUForm *response)
{
	ContinuationStateBuffer *pCStateBuffer;
	char *pdata;

	pdata = (char *)malloc(response->dataSize);
	if (!pdata)
		return -1;
	memcpy(pdata, response->data, response->dataSize);
	
	pCStateBuffer = (ContinuationStateBuffer *)malloc(sizeof(ContinuationStateBuffer));
	if (!pCStateBuffer) {
		free(pdata);
		return -1;
	}
	memset((char *)pCStateBuffer, 0, sizeof(ContinuationStateBuffer));

	pCStateBuffer->responseBuffer.data = pdata;
	pCStateBuffer->responseBuffer.dataSize = response->dataSize;
	pCStateBuffer->responseBuffer.bufferSize = response->dataSize;
	/* set owner */
	pCStateBuffer->fd = fd;
	/* set id */
	pCStateBuffer->timestamp = getTimeInSeconds();

	BTDEBUG("CacheList : 0x%lx", (long)cacheList);
	s_list_append(&cacheList, pCStateBuffer);

#ifdef CONFIG_AFFIX_DEBUG 
	s_list_foreach(cacheList, cacheListPrint, NULL);
#endif
	return pCStateBuffer->timestamp;
}

static int cacheCmp(const void *value1, const void *value2)
{
	ContinuationStateBuffer *csb1 = (ContinuationStateBuffer*)value1;
	ContinuationStateBuffer *csb2 = (ContinuationStateBuffer*)value2;

	if (!csb1 || !csb2)
		return -1;

	if (csb1->fd == csb2->fd && 
			(!csb2->timestamp || csb1->timestamp == csb2->timestamp))
		return 0;
	return -1;
}

PDUForm *getCachedResponse(int fd, ContinuationState *pCState)
{
	PDUForm			*pduForm;
	slist_t 		*entry;
	ContinuationStateBuffer *realRec;
	ContinuationStateBuffer toFind;

#ifdef CONFIG_AFFIX_DEBUG
	s_list_foreach(cacheList, cacheListPrint, NULL);
#endif
	toFind.fd = fd;
	toFind.timestamp = pCState->timestamp;
	entry = s_list_find_custom(cacheList, &toFind, cacheCmp);
	if (!entry)
		return NULL;

	realRec = (ContinuationStateBuffer*)entry->data;
	if (!realRec)
		return NULL;

	pduForm = &realRec->responseBuffer;
	BTDEBUG("Cached response : 0x%lx", (long)pduForm);
	return pduForm;
}

static int cacheRemove(const void *value1, const void *value2)
{
	if (!value1)
		return -1;
	if (!value2 || cacheCmp(value1, value2) == 0) {
		ContinuationStateBuffer *pBuffer = (ContinuationStateBuffer*)value1;

		free(pBuffer->responseBuffer.data);
		free(pBuffer);
		return 0;
	}
	return -1;
}

void removeCachedResponse(int fd, ContinuationState *pCState)
{
	ContinuationStateBuffer toRemove;

	toRemove.fd = fd;
	toRemove.timestamp = pCState ? pCState->timestamp : 0;
	s_list_remove_custom(&cacheList, &toRemove, cacheRemove);
}

void cleanupContinuationStateBuffer(void)
{
	BTDEBUG("Cache table cleanup");
	s_list_remove_custom(&cacheList, NULL, cacheRemove);
	
}

int copyContinuationState(char *pdata, CState *pCState)
{
	int length;

	if (pCState == NULL) {
		__put_u8(pdata, 0);
		length = 1;
	} else {
		__put_u8(pdata++, pCState->length);
		memcpy(pdata, pCState->data, pCState->length);
		length = pCState->length+1;
	}
	return length;
}

