/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001,2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Original Author: Guruprasad Krishnamurthy <kgprasad@hotmail.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: des.c,v 1.14 2002/11/04 15:29:58 kds Exp $

   Extracting attributes from received PDU

   Fixes:
   		Manel Guerrero Zapata <manel.guerrero-zapata@nokia.com>
		Dmitry Kasatkin
*/

#include <affix/config.h>

#include <stdint.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include "utils.h"
#include "attr.h"


SDPData *extractAttribute(char *pdata, int *extractedLength, ServiceRecord *svcRec);


/*
 ** Extract the sequence type, it's length
 ** and return a pointer to the beginning of
 ** sequence
 */
char *extractSequenceType(
		char *dstBuffer,
		DataTypeDescriptor *dataType,
		int *sequenceSize,
		int *bytesScanned
		)
{
	char *pdata;
	DataTypeDescriptor localDataType;

	pdata = dstBuffer;
	localDataType = __get_u8(pdata);
	pdata += sizeof(DataTypeDescriptor);
	*bytesScanned += sizeof(DataTypeDescriptor);
	*dataType = localDataType;

	switch (localDataType) {
		case SequenceSize8 :
		case AlternativeSize8 :
			*sequenceSize = __get_u8(pdata);
			pdata += sizeof(DataSizeDescriptor8);
			*bytesScanned += sizeof(DataSizeDescriptor8);
			break;

		case SequenceSize16 :
		case AlternativeSize16 :
			*sequenceSize = ntohs(__get_u16(pdata));
			pdata += sizeof(DataSizeDescriptor16);
			*bytesScanned += sizeof(DataSizeDescriptor16);
			break;

		case SequenceSize32 :
		case AlternativeSize32 :
			*sequenceSize = ntohl(__get_u32(pdata));
			pdata += sizeof(DataSizeDescriptor32);
			*bytesScanned += sizeof(DataSizeDescriptor32);
			break;

		default :
			BTERROR("Unknown sequence type, aborting\n");
			pdata = NULL;
			break;
	}
	return pdata;
}

/*
 ** Extract a sequence of service record handles from a PDU buffer
 ** and add the entries to a slist_t. Note that the service record
 ** handles are not in "data element sequence" form, but just like
 ** an array of service handles
 */
int extractServiceRecordHandles(
		char *pduBuffer,
		slist_t **svcReqSeq,
		int currentHandleCount,
		int *bytesScanned
		)
{
	int			status = -1;
	char			*pdata = pduBuffer;
	ServiceRecordHandle	*pSvcRec;
	int 			numberOfSvcRec = 0;

	for (;;) {
		pSvcRec = (ServiceRecordHandle *)malloc(sizeof(ServiceRecordHandle));
		*pSvcRec = ntohl(__get_u32(pdata));
		s_list_append(svcReqSeq, pSvcRec);
		pdata += sizeof(ServiceRecordHandle);
		*bytesScanned += sizeof(ServiceRecordHandle);
		numberOfSvcRec++;
		if (numberOfSvcRec == currentHandleCount) {
			status = 0;
			break;
		}
	}
	return status;
}

SDPData *extractInteger(char *pdata, int *extractedLength)
{
	SDPData *pIntData = NULL;
	DataTypeDescriptor dtd;

	BTDEBUG("Extracting integer");
	dtd = __get_u8(pdata);
	pdata += sizeof(DataTypeDescriptor);
	*extractedLength += sizeof(DataTypeDescriptor);

	pIntData = (SDPData *)malloc(sizeof(SDPData));
	memset(pIntData, 0, sizeof(SDPData));
	switch (dtd) {
		case Boolean :
		case Integer_8 :
		case UnsignedInteger_8 :
			*extractedLength += sizeof(uint8_t);
			pIntData->sdpDataValue.uint8 = __get_u8(pdata);
			break;
		case Integer_16 :
		case UnsignedInteger_16 :
			*extractedLength += sizeof(uint16_t);
			pIntData->sdpDataValue.uint16 = ntohs(__get_u16(pdata));
			break;
		case Integer_32 :
		case UnsignedInteger_32 :
			*extractedLength += sizeof(uint32_t);
			pIntData->sdpDataValue.uint32 = ntohl(__get_u32(pdata));
			break;
			/*
			 ** Not clear if NBO rules apply here and if yes, how
			 */
		case Integer_64 :
		case UnsignedInteger_64 :
			*extractedLength += sizeof(uint64_t);
			break;
		case Integer_128 :
		case UnsignedInteger_128 :
			*extractedLength += sizeof(Int128Bit);
			break;
		default :
			free(pIntData);
			pIntData = NULL;
	}

#ifdef CONFIG_AFFIX_DEBUG
	if (pIntData != NULL)
		BTDEBUG("Ingeter : 0x%x", pIntData->sdpDataValue.uint32);
#endif

	return pIntData;
}

SDPData *extractUUIDRec(char *pdata, int *extractedLength, ServiceRecord *svcRec)
{
	int status = 0;
	SDPData *pUUIDData = NULL;

	BTDEBUG("Extracting UUID");
	pUUIDData = (SDPData *)malloc(sizeof(SDPData));
	memset(pUUIDData, 0, sizeof(SDPData));
	status = extractUUID(pdata, &pUUIDData->sdpDataValue.uuid, extractedLength);
	if (status == -1) {
		free(pUUIDData); pUUIDData = NULL;
	} else
		addUUIDToTargetPattern(svcRec, &pUUIDData->sdpDataValue.uuid);

	return pUUIDData;
}

/*
 ** Extract strings from the PDU (could be service description
 ** and similar info) 
 */
int extractStringInt(char *buffer, char **dstStringBuffer, int *bytesScanned)
{
	char *pdata = buffer;
	int stringLength = 0;
	DataTypeDescriptor dataType;

	dataType = __get_u8(pdata);
	pdata += sizeof(DataTypeDescriptor);
	*bytesScanned += sizeof(DataTypeDescriptor);

	switch (dataType) {
		case TextStringSize8 :
		case URLStringSize8 :
			stringLength = __get_u8(pdata);
			pdata += sizeof(DataSizeDescriptor8);
			*bytesScanned += sizeof(DataSizeDescriptor8) + stringLength;
			break;

		case TextStringSize16 :
		case URLStringSize16 :
			stringLength = ntohs(__get_u16(pdata));
			pdata += sizeof(DataSizeDescriptor16);
			*bytesScanned += sizeof(DataSizeDescriptor16) + stringLength;
			break;

		default :
			BTERROR("Sizeof text string > UINT16_MAX");
			return -1;
	}

	/*
	 ** Extract the string
	 */
	*dstStringBuffer = (char *)malloc(stringLength + 1);
	memset(*dstStringBuffer, 0, (stringLength + 1));
	strncpy(*dstStringBuffer, pdata, stringLength);
	BTDEBUG("Len : %d", stringLength);
	BTDEBUG("Str : %s", *dstStringBuffer);

	return 0;
}

SDPData *extractString(char *pdata, int *extractedLength)
{
	int status = 0;
	SDPData *pStringData = NULL;

	BTDEBUG("Extracting string");
	pStringData = (SDPData *)malloc(sizeof(SDPData));
	memset(pStringData, 0, sizeof(SDPData));
	status = extractStringInt(pdata, &pStringData->sdpDataValue.stringPtr, 
			extractedLength);
	if (status == -1) {
		free(pStringData); pStringData = NULL;
	}

	return pStringData;
}

SDPData *extractSequence(char *pdata, int *extractedLength, ServiceRecord *svcRec)
{
	int		sequenceLength = 0;
	int		localExtractedLength = 0;
	SDPData		*currentNode;
	int		seqElementCount = 0;
	char		*localSeqPtr;
	SDPData		*pDataSeq;
	int		attributeLength = 0;
	slist_t		*dataSeq = NULL;
	DataTypeDescriptor dtd;

	BTDEBUG("Extracting sequence ");
	localSeqPtr = extractSequenceType(pdata, &dtd, &sequenceLength, extractedLength);
	BTDEBUG("Sequence Type : 0x%x length : 0x%x", dtd, sequenceLength);

	if ((localSeqPtr == NULL) || (sequenceLength == 0))
		return NULL;

	while (localExtractedLength < sequenceLength) {
		attributeLength = 0;
		currentNode = extractAttribute(localSeqPtr, &attributeLength, svcRec);
		if (currentNode == NULL)
			break;

		seqElementCount++;
		s_list_append(&dataSeq, currentNode);
		localSeqPtr += attributeLength;
		localExtractedLength += attributeLength;
		BTDEBUG("LocalExtractedLength : %d SequenceLength : %d", localExtractedLength, sequenceLength);
	}
	if (localExtractedLength != sequenceLength) {
		BTDEBUG("Error in extracting sequence, terminating");
		freeSDPDataSeq(dataSeq);
		return NULL;
	}

	pDataSeq = malloc(sizeof(SDPData));
	if (pDataSeq == NULL) {
		freeSDPDataSeq(dataSeq);
		return NULL;
	}
	memset(pDataSeq, 0, sizeof(SDPData));
	pDataSeq->sdpDataValue.dataSeq = dataSeq;
			
	*extractedLength += localExtractedLength;

	return pDataSeq;
}


SDPData *extractAttribute(char *pdata, int *extractedLength, ServiceRecord *svcRec)
{
	DataTypeDescriptor dataType;
	SDPData *pDataElem = NULL;
	int localExtractedLength = 0;

	dataType = __get_u8(pdata);

	BTDEBUG("DataTypeDescriptor : 0x%x", dataType);

	switch (dataType) {
		case Boolean :
		case UnsignedInteger_8 :
		case UnsignedInteger_16 :
		case UnsignedInteger_32 :
		case UnsignedInteger_64 :
		case UnsignedInteger_128 :
		case Integer_8 :
		case Integer_16 :
		case Integer_32 :
		case Integer_64 :
		case Integer_128 :
			pDataElem = extractInteger(pdata, &localExtractedLength);
			break;

		case UUID_16 :
		case UUID_32 :
		case UUID_128 :
			pDataElem = extractUUIDRec(pdata, &localExtractedLength, svcRec);
			break;

		case TextStringSize8 :
		case TextStringSize16 :
		case TextStringSize32 :
		case URLStringSize8 :
		case URLStringSize16 :
		case URLStringSize32 :
			pDataElem = extractString(pdata, &localExtractedLength);
			break;

		case SequenceSize8 :
		case SequenceSize16 :
		case SequenceSize32 :
			pDataElem = extractSequence(pdata, &localExtractedLength, svcRec);
			break;

		case AlternativeSize8 :
		case AlternativeSize16 :
		case AlternativeSize32 :
			pDataElem = extractSequence(pdata, &localExtractedLength, svcRec);
			break;

		default :
			BTERROR("Unknown data descriptor : 0x%x terminating", dataType);
			free(pDataElem);
			pDataElem = NULL;
			break;
	}

	if (pDataElem != NULL) {
		pDataElem->dtd = dataType;
		*extractedLength += localExtractedLength;
	}

	return pDataElem;
}

int extractAttributes(ServiceRecord *svcRec, char *localSeqPtr, int sequenceLength)
{
	int		extractStatus = -1;
	int		localExtractedLength = 0;
	DataTypeDescriptor dtd;
	int		attrValueLength;// = 0;
	int		attrSize;// = 0;
	ServiceAttributeIdentifier attrId;
	SDPData		*pAttrValue;// = NULL;

	while (localExtractedLength < sequenceLength) {
		BTDEBUG("Extract PDU, sequenceLength : %d localExtractedLength : %d",
				sequenceLength, localExtractedLength);
		attrSize = 0;
		dtd = __get_u8(localSeqPtr);
		attrSize = sizeof(DataTypeDescriptor);
		attrId = ntohs(__get_u16(localSeqPtr + attrSize));
		attrSize += sizeof(ServiceAttributeIdentifier);
		BTDEBUG("DTD of attrId : %d Attr id : 0x%x ", dtd, attrId);
		attrValueLength = 0;
		pAttrValue = extractAttribute((localSeqPtr+attrSize), &attrValueLength, svcRec);
		BTDEBUG("Attr id : 0x%x attrValueLength : %d", attrId, attrValueLength);
		attrSize += attrValueLength;
		if (pAttrValue == NULL) {
			BTDEBUG("Terminating extraction of attributes");
			break;
		} else {
			pAttrValue->attrId = attrId;
			localExtractedLength += attrSize;
			localSeqPtr += attrSize;
			addToAttributeList(svcRec, attrId, pAttrValue, 1);
			extractStatus = 0;
			BTDEBUG("Extract PDU, sequenceLength : %d localExtractedLength : %d",
					sequenceLength, localExtractedLength);
		}
	}

	if (extractStatus == 0) {
		BTDEBUG("Successful extracting of Svc Rec attributes");
		printServiceAttributes(svcRec->attributeList);
		return sequenceLength;
	}
	
	return 0;
}


/* 
 * Extract PDU for Client
 */
ServiceRecord *clientExtractPDU(char *pdata, ServiceRecordHandle handleExpected, int *bytesScanned)
{
	int localExtractedLength = 0;
	char *localSeqPtr = NULL;
	DataTypeDescriptor dtd;
	int sequenceLength = 0;
	ServiceRecord *svcRec = NULL;
	ServiceAttributeIdentifier lookAheadAttrId = 0xffff;
	ServiceRecordHandle svcRecHandle = 0xffffffff;

	localSeqPtr = extractSequenceType(pdata, &dtd, &sequenceLength, 
			bytesScanned);
	lookAheadAttrId = ntohs(__get_u16(localSeqPtr + sizeof(DataTypeDescriptor)));
	BTDEBUG("Look ahead attr id : %d", lookAheadAttrId);

	/*
	 ** Check if the PDU already has a service record handle,
	 ** int which case this may be a update request
	 */
	if (lookAheadAttrId == AttrServiceRecordHandle) {
		svcRecHandle = ntohl(__get_u32(localSeqPtr + 
					sizeof(DataTypeDescriptor) +
					sizeof(ServiceAttributeIdentifier) +
					sizeof(DataTypeDescriptor)));
		BTDEBUG("SvcRecHandle : 0x%x", svcRecHandle);
	}

	svcRec = createServiceRecord();
	if (svcRec == NULL)
		return NULL;
	svcRec->attributeList = NULL;
	if (lookAheadAttrId == AttrServiceRecordHandle) {
		svcRec->serviceRecordHandle = svcRecHandle;
	} else {
		if (handleExpected != 0xffffffff) {
			svcRec->serviceRecordHandle = handleExpected;
		} else {
			BTINFO("ServiceRecordHandle not defined");
		}
	}

	localExtractedLength = extractAttributes(svcRec, localSeqPtr, sequenceLength);
	if (localExtractedLength >= 0)
		*bytesScanned += localExtractedLength;

	return svcRec;
}

