/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001,2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Original Author: Guruprasad Krishnamurthy <kgprasad@hotmail.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
   $Id: sdpclt.c,v 1.21 2002/11/26 17:05:05 kds Exp $

   Contains the implementation of the SDP service discovery API
 
   Fixes:

   		Dmitry Kasatkin		- Continuation Mechanism
		Dmitry Kasatkin		- Port Request
		Dmitry Kasatkin		- cleaning

*/
	 
#include <affix/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include "utils.h"
#include "cstate.h"
#include "attr.h"

/*
** This is a service search request. 
**
** INPUT :
**
**   slist_t *svcSearchList
**     Singly linked list containing elements of the search
**     pattern. Each entry in the list is a UUID (DataTypeUUID_16)
**     of the service to be searched
**
**   uint16_t maxSvcRecordCount
**      A 16 bit integer which tells the service, the maximum
**      entries that the client can handle in the response. The
**      server is obliged not to return > maxSvcRecordCount entries
**
** OUTPUT :
**
**   int return value
**     E_OK 
**       The request completed successfully. This does not
**       mean the requested services were found
**     E_FAILURE
**       On any failure
**     E_TIMEOUT 
**       The request completed unsuccessfully due to a timeout
**
**   slist_t **svcHandleList
**     This variable is set on a successful return if there are
**     non-zero service handles. It is a singly linked list of
**     service record handles (ServiceRecordHandle)
**
**   uint16_t *handleCount
**     This is a pointer to a 16 bit integer, which is set to 
**     indicate the number of service record handles present in
**     svcHandleList
*/
int SDPServiceSearchRequest(
		SDPServerHandle srvHandle,
		slist_t *svcSearchList, 
		uint16_t maxSvcRecordCount,
		slist_t **svcResponseList,
		uint16_t *handleCountInResponse)
{
	int status = 0;
	int requestSize = 0, _requestSize;
	int responseSize = 0;
	int seqLength = 0;
	char *pdata, *_pdata;
	char *requestBuffer =  NULL;
	char *responseBuffer =  NULL;
	SdpPduHeader *pduRequestHeader;
	SdpPduHeader *pduResponseHeader;
	CState *pCState = NULL;
	//PDUForm concatenatedResponseBuffer;
	int responseLength = 0;

	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceSearchRequest;
	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);

	/*
	 ** Add service class IDs for search
	 */
	seqLength = generateSearchSequencePduForm(pdata, svcSearchList);

	BTDEBUG("Data seq added : %d\n", seqLength);

	/*
	 ** Now set the length and increment the pointer
	 */
	requestSize += seqLength;
	pdata += seqLength;

	/*
	 ** Specify the maximum svc rec count that
	 ** client expects
	 */
	__put_u16(pdata, htons(maxSvcRecordCount));
	requestSize += sizeof(uint16_t);
	pdata += sizeof(uint16_t);

	_requestSize = requestSize;
	_pdata = pdata;
	*svcResponseList = NULL;	//important

	do {
		int bytesScanned = 0;
		int totalServiceRecordCount, currentServiceRecordCount;

		/*
		 ** Add continuation state or NULL (first time)
		 */
		requestSize = _requestSize + copyContinuationState(_pdata, pCState);

		/*
		 ** Set the request header's param length
		 */
		pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));
		pduRequestHeader->transactionId = htons(generateTransactionId());
		/*
		 ** Send the request, wait for response and if
		 ** not error, set the appropriate values
		 ** and return
		 */
		status = sendRequestWaitForResponse(srvHandle, requestBuffer, 
				responseBuffer, requestSize, &responseSize);

		if (status != E_OK)
			break;

		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		if (pduResponseHeader->pduId == SDP_ErrorResponse) {
			status = E_FAILURE;
			break;
		}

		responseLength = ntohs(pduResponseHeader->paramLength);
		pdata = responseBuffer + sizeof(SdpPduHeader);
		/*
		 ** Net service record match count
		 */
		totalServiceRecordCount = 0;
		currentServiceRecordCount = 0;
		totalServiceRecordCount = ntohs(__get_u16(pdata));
		pdata += sizeof(uint16_t);
		bytesScanned += sizeof(uint16_t);
		currentServiceRecordCount = ntohs(__get_u16(pdata));
		pdata += sizeof(uint16_t);
		bytesScanned += sizeof(uint16_t);

		BTDEBUG("Total svc count : %d\n", totalServiceRecordCount);
		BTDEBUG("Current svc count : %d\n", currentServiceRecordCount);
		BTDEBUG("ResponseLength : %d\n", responseLength);
		BTDEBUG("Total svc count : %d\n", totalServiceRecordCount);
		BTDEBUG("Current svc count : %d\n", currentServiceRecordCount);
		BTDEBUG("ResponseLength : %d\n", responseLength);

		if (currentServiceRecordCount == 0) {
			status = E_OK;
			break;
		}
		status = extractServiceRecordHandles(pdata, svcResponseList, 
				currentServiceRecordCount, &bytesScanned);
		if (status)
			break;
		
		*handleCountInResponse = currentServiceRecordCount;
		/*
		 ** Check if we have ContinuationState set, if yes,
		 ** set the pCState pointer
		 */
		BTDEBUG("BytesScanned : %d\n", bytesScanned);
		if (responseLength > bytesScanned) {
			uint8_t cStateLength = 0;

			pdata = responseBuffer + sizeof(SdpPduHeader) + bytesScanned;
			cStateLength = __get_u8(pdata);
			if (cStateLength > 0) {
				pCState = (CState *)pdata;
				BTDEBUG("Continuation state present\n");
				BTDEBUG("CState length : %d\n", cStateLength);
			} else
				pCState = NULL;
		}
	} while (pCState);
	
	free(requestBuffer);
	free(responseBuffer);
	//if (concatenatedResponseBuffer.data != NULL)
	//free(concatenatedResponseBuffer.data);

	return status;
}

int __SDPServiceSearchRequest(
		struct sockaddr_affix *sa,
		slist_t *svcSearchList, 
		uint16_t maxSvcRecordCount,
		slist_t **svcResponseList,
		uint16_t *handleCountInResponse)
{
	int status = 0;
	SDPServerHandle srvHandle;

	srvHandle = SDPOpenServerConnection(sa);
	if (srvHandle < 0)
		return srvHandle;

	status  = SDPServiceSearchRequest(srvHandle, svcSearchList, 
			maxSvcRecordCount, svcResponseList, handleCountInResponse);
	SDPCloseServerConnection(srvHandle);

	return status;
}



/*
** This is a service attribute request. 
**
** INPUT :
**
**   ServiceRecordHandle svcHandle
**     The handle of the service for which the attribute(s) are
**     requested
**
**   AttributeRequestType attrReqType
**     Attribute identifiers are 16 bit unsigned integers specified
**     in one of 2 ways described below :
**     IndividualAttributes - 16bit individual identifiers
**        They are the actual attribute identifiers in ascending order
**
**     RangeOfAttributes - 32bit identifier range
**        The high-order 16bits is the start of range
**        the low-order 16bits are the end of range
**        0x0000 to 0xFFFF gets all attributes
**
**   slist_t *attrIDList
**     Singly linked list containing attribute identifiers desired.
**     Every element is either a uint16_t(attrSpec = IndividualAttributes)  
**     or a uint32_t(attrSpec=RangeOfAttributes)
**
**   uint16_t maxAttrIDByteCount
**     The byte count of the number of attribute IDs specified
**     in the request list
**
** OUTPUT :
**   int return value
**     E_OK 
**       The request completed successfully. This does not
**       mean the requested services were found
**     E_TIMEOUT 
**       The request completed unsuccessfully due to a timeout
**
**   uint16_t *maxAttrResponseByteCount
**     This is a pointer to a 16 bit integer, which is set to 
**     indicate the number of bytes of attributes returned. 
**     This pointer is set on successful return
**
*/


int SDPServiceAttributeRequest(
		SDPServerHandle srvHandle,
		ServiceRecordHandle svcHandle,
		AttributeRequestType attrReqType,
		slist_t *attrIDList,
		uint16_t maxAttrIDByteCount,
		ServiceRecord	**_svcRec,
		uint16_t *maxAttrResponseByteCount)
{
	int status = 0;
	int requestSize = 0, _requestSize;
	int responseSize = 0;
	int attrListByteCount = 0;
	int seqLength = 0;
	char *pdata = NULL, *_pdata;
	char *requestBuffer =  NULL;
	char *responseBuffer =  NULL;
	SdpPduHeader *pduRequestHeader;
	SdpPduHeader *pduResponseHeader;
	// DataTypeDescriptor dataType = 0;
	CState *pCState = NULL;
	uint8_t cStateLength = 0;
	PDUForm concatenatedResponseBuffer;
	ServiceRecord *svcRec = NULL;

	*_svcRec = NULL;

	if ((attrReqType != IndividualAttributes) && (attrReqType != RangeOfAttributes))
		return E_FAILURE;

	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	memset((char *)&concatenatedResponseBuffer, 0, sizeof(PDUForm));

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceAttributeRequest;

	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);

	/*
	 ** Add the service record handle
	 */
	__put_u32(pdata, htonl(svcHandle));
	requestSize += sizeof(ServiceRecordHandle);
	pdata += sizeof(ServiceRecordHandle);

	/*
	 ** Add the maxAttrIDByteCount specifying response limit
	 */
	__put_u16(pdata, htons(maxAttrIDByteCount));
	requestSize += sizeof(uint16_t);
	pdata += sizeof(uint16_t);

	/*
	 ** Get attr seq PDU form 
	 */
	seqLength = generateAttrIDSequencePduForm(pdata, attrIDList, 
			((attrReqType == IndividualAttributes) ? UnsignedInteger_16 : UnsignedInteger_32));
	if (seqLength == -1) {
		status = E_INVALID_ARGS;
		goto exit;
	} 

	pdata += seqLength;
	requestSize += seqLength;
	BTDEBUG("Attr list length : %d\n", seqLength);

	// save before Continuation State
	_pdata = pdata;
	_requestSize = requestSize;

	do {
		int responseCount = 0;
		/*
		 ** Add NULL continuation state
		 */
		requestSize = _requestSize + copyContinuationState(_pdata, pCState);

		/*
		 ** Set the request header's param length
		 */
		pduRequestHeader->transactionId = htons(generateTransactionId());
		pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));

		status = sendRequestWaitForResponse(srvHandle, requestBuffer, 
				responseBuffer, requestSize, &responseSize);

		if (status != E_OK)
			goto exit;

		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		if (pduResponseHeader->pduId == SDP_ErrorResponse) {
			status = E_FAILURE;
			goto exit;
		}

		pdata = responseBuffer + sizeof(SdpPduHeader);
		responseCount  = ntohs(__get_u16(pdata));
		attrListByteCount += responseCount;
		pdata += sizeof(uint16_t);

		/*
		 ** Check if we have continuation state set, if yes, we
		 ** need to re-issue request before we parse ..
		 */
		cStateLength = __get_u8(pdata + responseCount);
		BTDEBUG("Response id : %d\n", pduResponseHeader->pduId);
		BTDEBUG("Attrlist byte count : %d\n", responseCount);
		BTDEBUG("CState length : %d\n", cStateLength);
		/*
		 ** This is a split response, need to concatenate the intermediate
		 ** responses as well as the last one which will have cStateLength == 0
		 */
		if ((cStateLength > 0) || (concatenatedResponseBuffer.dataSize != 0)) {
			char *targetPtr = NULL;

			if (cStateLength > 0)
				pCState = (CState *)(pdata + responseCount);
			else
				pCState = NULL;

			/*
			 ** Build concatenated response buffer
			 */
			concatenatedResponseBuffer.data = (char *)realloc(
					concatenatedResponseBuffer.data, 
					concatenatedResponseBuffer.dataSize + responseCount);
			concatenatedResponseBuffer.bufferSize = 
				concatenatedResponseBuffer.dataSize + responseCount;
			targetPtr = concatenatedResponseBuffer.data + 
				concatenatedResponseBuffer.dataSize;
			memcpy(targetPtr, pdata, responseCount);
			concatenatedResponseBuffer.dataSize += responseCount;
		}
	} while (pCState);

	if (attrListByteCount > 0) {
		int bytesScanned = 0;

		if (concatenatedResponseBuffer.dataSize != 0) {
			pdata = concatenatedResponseBuffer.data;
		}
		svcRec = clientExtractPDU(pdata, svcHandle, &bytesScanned);
		BTDEBUG("In SDPServiceAttributeRequest function \n");
		BTDEBUG("The status of extractServiceRecord is %d\n", status);
		BTDEBUG("The svcHandle is 0x%x\n", svcHandle);
		if (svcRec != NULL) {
			*_svcRec = svcRec;
			*maxAttrResponseByteCount = bytesScanned;
		} else
			status = E_FAILURE;
	}
exit:
	free(requestBuffer);
	free(responseBuffer);
	if (concatenatedResponseBuffer.data != NULL)
		free(concatenatedResponseBuffer.data);
	return status;
}

int __SDPServiceAttributeRequest(
		struct sockaddr_affix *sa,
		ServiceRecordHandle svcHandle,
		AttributeRequestType attrReqType,
		slist_t *attrIDList,
		uint16_t maxAttrIDByteCount,
		ServiceRecord	**svcRec,
		uint16_t *maxAttrResponseByteCount)
{
	int status = 0;
	SDPServerHandle srvHandle;

	srvHandle = SDPOpenServerConnection(sa);
	if (srvHandle < 0)
		return srvHandle;

	status = SDPServiceAttributeRequest(srvHandle, svcHandle, attrReqType,
			attrIDList, maxAttrIDByteCount, svcRec, maxAttrResponseByteCount);
	SDPCloseServerConnection(srvHandle);

	return status;
}


/*
** This is a service search request combined with the service
** attribute request. First a service class match is done and
** for matching service, requested attributes are extracted
**
** INPUT :
**
**   slist_t *svcSearchList
**     Singly linked list containing elements of the search
**     pattern. Each entry in the list is a UUID(DataTypeUUID_16)
**     of the service to be searched
**
**   AttributeSpecification attrSpec
**     Attribute identifiers are 16 bit unsigned integers specified
**     in one of 2 ways described below :
**     IndividualAttributes - 16bit individual identifiers
**        They are the actual attribute identifiers in ascending order
**
**     RangeOfAttributes - 32bit identifier range
**        The high-order 16bits is the start of range
**        the low-order 16bits are the end of range
**        0x0000 to 0xFFFF gets all attributes
**
**   slist_t *attrIDList
**     Singly linked list containing attribute identifiers desired.
**     Every element is either a uint16_t(attrSpec = IndividualAttributes)  
**     or a uint32_t(attrSpec=RangeOfAttributes)
**
**   uint16_t maxAttrIDByteCount
**     The byte count of the number of attribute IDs specified
**     in the request list
**
** OUTPUT :
**   int return value
**     E_OK 
**       The request completed successfully. This does not
**       mean the requested services were found
**     E_TIMEOUT 
**       The request completed unsuccessfully due to a timeout
**
**   slist_t **svcResponseList
**     This variable is set on a successful return to point to
**     service(s) found. Each element of this list is of type
**     ServiceRecordHandle (of the services which matched the search list)
**
**   uint16_t *maxAttrResponseByteCount
**     This is a pointer to a 16 bit integer, which is set to 
**     indicate the number of bytes of attributes returned. 
**     This pointer is set on successful return
**
*/
int SDPServiceSearchAttributeRequest(
		SDPServerHandle srvHandle,
		slist_t *svcSearchList, 
		AttributeRequestType attrReqType,
		slist_t *attrIDList,
		uint16_t maxAttrByteCount,
		slist_t **svcResponseList,
		uint16_t *maxAttrResponseByteCount
		)
{
	int status = 0;
	int requestSize = 0, _requestSize;
	int responseSize = 0;
	int seqLength = 0;
	int attrListByteCount = 0;
	char *pdata = NULL, *_pdata;
	char *requestBuffer =  NULL;
	char *responseBuffer =  NULL;
	SdpPduHeader *pduRequestHeader;
	SdpPduHeader *pduResponseHeader;
	DataTypeDescriptor dataType;
	slist_t *svcRecHandleList = NULL;
	PDUForm concatenatedResponseBuffer;
	CState	*pCState = NULL;
	int bytesScanned = 0;

	if ((attrReqType != IndividualAttributes) && (attrReqType != RangeOfAttributes))
		return E_FAILURE;
	
	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	memset((char *)&concatenatedResponseBuffer, 0, sizeof(PDUForm));

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceSearchAttributeRequest;

	// generate PDU
	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);

	/*
	 ** Add service class IDs for search
	 */
	seqLength = generateSearchSequencePduForm(pdata, svcSearchList);

	BTDEBUG("Data seq added : %d\n", seqLength);

	/*
	 ** Now set the length and increment the pointer
	 */
	requestSize += seqLength;
	pdata += seqLength;

	/*
	 ** Specify the maximum attr count that client expects
	 */
	__put_u16(pdata, htons(maxAttrByteCount));
	requestSize += sizeof(uint16_t);
	pdata += sizeof(uint16_t);

	BTDEBUG("Max attr byte count : %d\n", maxAttrByteCount);

	/*
	 ** Get attr seq PDU form 
	 */
	seqLength = generateAttrIDSequencePduForm(pdata, attrIDList, 
			((attrReqType == IndividualAttributes) ? UnsignedInteger_16 : UnsignedInteger_32));
	if (seqLength == -1) {
		status = E_INVALID_ARGS;
		goto exit;
	}
	pdata += seqLength;
	BTDEBUG("Attr list length : %d\n", seqLength);
	requestSize += seqLength;

	// save before Continuation State
	_pdata = pdata;
	_requestSize = requestSize;

	do {
		int responseCount = 0;
		int cStateLength = 0;

		pduRequestHeader->transactionId = htons(generateTransactionId());
		/*
		 * Add continuation state
		 * pCState can be null
		 */
		requestSize = _requestSize + copyContinuationState(_pdata, pCState);

		/*
		 ** Set the request header's param length
		 */
		pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));

		status = sendRequestWaitForResponse(srvHandle, requestBuffer, 
				responseBuffer, requestSize, &responseSize);

		if (status != 0)
			goto exit;

		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		if (pduResponseHeader->pduId == SDP_ErrorResponse) {
			status = E_FAILURE;
			goto exit;
		}

		pdata = responseBuffer + sizeof(SdpPduHeader);
		responseCount = ntohs(__get_u16(pdata));
		attrListByteCount += responseCount;
		pdata += sizeof(uint16_t);	// pdata points to attribute list

		/*
		 ** Check if we have continuation state set, if yes, we
		 ** need to re-issue request before we parse ..
		 */
		cStateLength = __get_u8(pdata + responseCount);
		BTDEBUG("Attrlist byte count : %d\n", attrListByteCount);
		BTDEBUG("Response byte count : %d\n", responseCount);
		BTDEBUG("Cstate length : %d\n", cStateLength);
		/*
		 ** This is a split response, need to concatenate the intermediate
		 ** responses as well as the last one which will have cStateLength == 0
		 */
		if ((cStateLength > 0) || (concatenatedResponseBuffer.dataSize != 0)) {
			char *targetPtr = NULL;

			if (cStateLength > 0) 
				pCState = (CState *)(pdata + responseCount);
			else
				pCState = NULL;
			/*
			 ** Build concatenated response buffer
			 */
			concatenatedResponseBuffer.data = (char *)realloc(
					concatenatedResponseBuffer.data, 
					concatenatedResponseBuffer.dataSize + responseCount);
			targetPtr = concatenatedResponseBuffer.data + 
				concatenatedResponseBuffer.dataSize;
			concatenatedResponseBuffer.bufferSize = 
				concatenatedResponseBuffer.dataSize + responseCount;
			memcpy(targetPtr, pdata, responseCount);
			concatenatedResponseBuffer.dataSize += responseCount;
		}
	} while (pCState);

	if (attrListByteCount <= 0)
		goto exit;

	if (concatenatedResponseBuffer.dataSize != 0)
		pdata = concatenatedResponseBuffer.data;
	/*
	 ** Response is a sequence of sequence(s) for one or
	 ** more data element sequence(s) representing services
	 ** for which attributes are returned
	 */
	pdata = extractSequenceType(pdata, &dataType, &seqLength, &bytesScanned);

	BTDEBUG("Bytes scanned : %d\n", bytesScanned);
	BTDEBUG("Seq length : %d\n", seqLength);
	if ((pdata == NULL) || (seqLength == 0)) {
		*maxAttrResponseByteCount = 0;
		goto exit;
	}


	for (;;) {
		int localSeqSize;
		ServiceRecord *svcRec = NULL;

		if (pdata == NULL) {
			BTDEBUG("This may be a spec interpretation anomaly !");
			BTDEBUG("Expect sequence of sequence(s)\n");
			status = E_FAILURE;
			break;
		}

		localSeqSize = 0;
		svcRec = clientExtractPDU(pdata, 0xffffffff, &localSeqSize);
		if (svcRec == NULL) {
			BTERROR("SVC REC is null\n");
			status = E_FAILURE;
			break;
		}

		bytesScanned += localSeqSize;
		*maxAttrResponseByteCount += localSeqSize;
		pdata += localSeqSize;

		BTDEBUG("Loc seq length : %d\n", localSeqSize);
		BTDEBUG("Svc Rec Handle : 0x%x\n", svcRec->serviceRecordHandle);
		BTDEBUG("Bytes scanned : %d\n", bytesScanned);
		BTDEBUG("Attrlist byte count : %d\n", attrListByteCount);
		s_list_append(&svcRecHandleList, svcRec);

		if (attrListByteCount == bytesScanned) {
			BTDEBUG("Successful scan of service attr lists\n");
			*svcResponseList = svcRecHandleList;
			break;
		}
	}
exit:
	free(requestBuffer);
	free(responseBuffer);
	if (concatenatedResponseBuffer.data != NULL)
		free(concatenatedResponseBuffer.data);
	return status;
}

int __SDPServiceSearchAttributeRequest(
		struct sockaddr_affix *sa,
		slist_t *svcSearchList, 
		AttributeRequestType attrReqType,
		slist_t *attrIDList,
		uint16_t maxAttrByteCount,
		slist_t **svcResponseList,
		uint16_t *maxAttrResponseByteCount
		)
{
	int status = 0;
	SDPServerHandle srvHandle;

	srvHandle = SDPOpenServerConnection(sa);
	if (srvHandle < 0)
		return srvHandle;

	status = SDPServiceSearchAttributeRequest(srvHandle, svcSearchList, attrReqType, attrIDList,
			maxAttrByteCount, svcResponseList, maxAttrResponseByteCount);
	SDPCloseServerConnection(srvHandle);
	return status;
}


/* --------------------------------------------------------------------------*/

/*
   Contains utilities for use by SDP service discovery clients
   to extract individual attributes from a service record
   All utility functions below have a signature 
   int getXXX(ServiceRecordHandle svcRec, XXX *pXXX)
   The pointer pXXX is set to point at the attribute value if
   it is present in the service record and 0 is returned, else
   E_NOT_EXIST is returned
*/

slist_t *getProtocolDescriptorList(slist_t *protSeq)
{
	slist_t		*protDescSeq = NULL;
	SDPData 	*currSDPData, *currEntrySDPData;

	for (;protSeq; protSeq = s_list_next(protSeq)) {
		ProtocolDescriptor	*pDesc;
		slist_t			*paramSeq;

		currSDPData = protSeq->data;

		/* get protocol parameters */
		paramSeq = currSDPData->sdpDataValue.dataSeq;
		if (paramSeq == NULL)
			continue;
		currEntrySDPData = paramSeq->data;
		pDesc = (ProtocolDescriptor *)malloc(sizeof(ProtocolDescriptor));               
		bzero(pDesc, sizeof(ProtocolDescriptor));
		BTDEBUG("Getting protocol id : %x\n", currEntrySDPData->dtd);
		pDesc->protocolId = currEntrySDPData->sdpDataValue.uuid;
		printUUID(&pDesc->protocolId);
		paramSeq = s_list_next(paramSeq);
		if (paramSeq != NULL) {
			/* first param - port */
			currEntrySDPData= paramSeq->data;
			if (currEntrySDPData->dtd == UnsignedInteger_16) {
				pDesc->portNumber = currEntrySDPData->sdpDataValue.uint16;
				pDesc->portNumberSize = 2;
			} else {
				pDesc->portNumber = currEntrySDPData->sdpDataValue.uint8;
				pDesc->portNumberSize = 1;
			}
			BTDEBUG("Port : %d\n", pDesc->portNumber);
			paramSeq = s_list_next(paramSeq);
			if (paramSeq != NULL) {
				/* second param - version */
				currEntrySDPData= paramSeq->data;
				pDesc->versionNumber = currEntrySDPData->sdpDataValue.uint16;
				BTDEBUG("Vnum : %d\n", pDesc->versionNumber);
			}
		}
		s_list_append(&protDescSeq, pDesc);
	}

	return protDescSeq;
}

int getAccessProtocols(ServiceRecord *svcRec, AccessProtocols **pap)
{
	SDPData 	*pSDPData;
	AccessProtocols	*ap = NULL;
	slist_t		*protDescSeq;

	*pap = NULL;
	pSDPData = getSDPData(svcRec, AttrProtocolDescriptorList);
	if (pSDPData == NULL)
		return E_NOT_EXIST;

	ap = (AccessProtocols *) malloc(sizeof(AccessProtocols));
	ap->pSequence = NULL;

	BTDEBUG("AP type : 0%x\n", pSDPData->dtd);

	if (isAlternativeData(pSDPData)) {
		slist_t	*protSeq, *protDescSeq;
		for (protSeq = pSDPData->sdpDataValue.dataSeq; protSeq; protSeq = s_list_next(protSeq)) {

			protDescSeq = getProtocolDescriptorList(protSeq);
			if (protDescSeq)
				s_list_append(&ap->pSequence, protDescSeq);
		}
	} else {
		protDescSeq = getProtocolDescriptorList(pSDPData->sdpDataValue.dataSeq);
		if (protDescSeq)
			s_list_append(&ap->pSequence, protDescSeq);

	}

	*pap = ap;
	return 0 ;
}

int getServiceClassIDList(ServiceRecord *svcRec, slist_t **svcClassSeq)
{
	return GetUUIDSequenceAttribute(svcRec, AttrServiceClassIDList, svcClassSeq); 
}

int getBrowseGroupList(ServiceRecord *svcRec, slist_t **ppslist_t)
{
	return GetUUIDSequenceAttribute(svcRec, AttrBrowseGroupList, ppslist_t); 
}


int getProtocolDescList(ServiceRecord *svcRec, slist_t **ppslist_t)
{
	return GetUUIDSequenceAttribute(svcRec, AttrProtocolDescriptorList, ppslist_t); 
}


int getServiceID(ServiceRecord *svcRec, UUID *pUUID)
{
	GetBasicAttribute(AttrServiceID, pUUID, uuid);
}


int getGroupID(ServiceRecord *svcRec, UUID *pUUID)
{
	GetBasicAttribute(AttrGroupID, pUUID, uuid);
}


unsigned char isGroup(ServiceRecord *svcRec)
{
	unsigned char exists = 0;

	SDPData *pSDPData = getSDPData(svcRec, AttrGroupID);
	if ((pSDPData != NULL) && 
			(svcRec->serviceRecordHandle != SDPServerServiceRecordHandle)) {
		exists = 1;
	}
	return exists;
}


int getServiceRecordState(ServiceRecord *svcRec, ServiceRecordState *svcRecState)
{
	GetBasicAttribute(AttrServiceRecordState, svcRecState, uint32);
}


int getServiceAvailability(ServiceRecord *svcRec, ServiceAvailability *svcAvail)
{
	GetBasicAttribute(AttrServiceAvailability, svcAvail, uint8);
}


int getServiceInfoTimeToLive(ServiceRecord *svcRec, ServiceInfoTimeToLive *svcTTLInfo)
{
	GetBasicAttribute(AttrServiceInfoTimeToLive, svcTTLInfo, uint32);
}


int getLanguageAttributes(ServiceRecord *svcRec, slist_t **langSeq)
{
	return 0; 
}


int getProfileDescriptor(ServiceRecord *svcRec, slist_t **profDescSeq)
{
	int		status = E_NOT_EXIST;
	ProfileDescriptor *profDesc;
	SDPData		*pSDPData = NULL;
	slist_t		*profSeq;
	SDPData		*pDataSeq;

	*profDescSeq = NULL;
	pSDPData = getSDPData(svcRec, AttrProfileDescriptorList);
	if ((pSDPData == NULL) || (pSDPData->sdpDataValue.dataSeq == NULL))
		return status;

	for (profSeq = pSDPData->sdpDataValue.dataSeq; profSeq; profSeq = s_list_next(profSeq)) {
		SDPData *pUUID = NULL;
		SDPData *pVnum = NULL;
		slist_t	*paramSeq;

		pDataSeq = profSeq->data;
		/* get parameters */		
		paramSeq = pDataSeq->sdpDataValue.dataSeq;
		if (paramSeq == NULL)
			continue;
		pUUID = paramSeq->data;
		paramSeq = s_list_next(paramSeq);
		if (paramSeq == NULL)
			continue;
		pVnum = paramSeq->data;
		if ((pUUID != NULL) && (pVnum != NULL)) {
			profDesc = (ProfileDescriptor *)malloc(sizeof(ProfileDescriptor));
			profDesc->profileUUID =  pUUID->sdpDataValue.uuid;
			profDesc->vnum = pVnum->sdpDataValue.uint16;
			printUUID(&profDesc->profileUUID);
			BTDEBUG("Vnum : %d\n", profDesc->vnum);
			s_list_append(profDescSeq, profDesc);
			status = E_OK;
		}
	}

	return status;
}


int getServiceName(ServiceRecord *svcRec, char **pName)
{
	return GetStringAttribute(svcRec, AttrServiceName_Primary, pName);
}


int getServiceDescription(ServiceRecord *svcRec, char **pDesc)
{
	return GetStringAttribute(svcRec, AttrServiceDescription_Primary, pDesc);
}


int getServiceProviderName(ServiceRecord *svcRec, char **pSvcProviderName)
{
	return GetStringAttribute(svcRec, AttrServiceProviderName_Primary, pSvcProviderName);
}


int getDocumentationURL(ServiceRecord *svcRec, char **pDocURL)
{
	return GetStringAttribute(svcRec, AttrDocumentationURL, pDocURL);
}


int getClientExecutableURL(ServiceRecord *svcRec, char **pClientURL)
{
	return GetStringAttribute(svcRec, AttrClientExecutableURL, pClientURL);
}


int getIconURL(ServiceRecord *svcRec, char **pIconURL)
{
	return GetStringAttribute(svcRec, AttrIconURL, pIconURL);
}


void printServiceRecordById(ServiceRecord *svcRec)
{
	// printServiceRecord(bdAddr, svcRecHandle);
	printf("printServiceRecordById: Print the ServiceRecord\n");
	// printPduForm(&svcRec->pduForm);
	generateServiceRecordPduForm(svcRec);
	printSDPDataAsServiceRecord(svcRec);

}


int getSDPServerServiceDatabaseState(ServiceRecord *svcRec, ServiceDatabaseState *svcDBState)
{
	GetBasicAttribute(AttrServiceDatabaseState, svcDBState, uint32);
}


int getSDPServerVersionNumber(ServiceRecord *svcRec, slist_t **pVersionNumberSeq)
{
	int status = E_NOT_EXIST;
	SDPData *pSDPData, *currSDPData;
	slist_t	*list;

	*pVersionNumberSeq = NULL;
	pSDPData = getSDPData(svcRec, AttrVersionNumberList);
	if (pSDPData != NULL) {
		for (list = pSDPData->sdpDataValue.dataSeq; list != NULL; list = s_list_next(list)) { 
			currSDPData = list->data;
			s_list_append(pVersionNumberSeq, &currSDPData->sdpDataValue.uint16);
		}
		status = E_OK;
	}
	return status;
}

/* -------------------------------  new ------------------------ */

int getRFCOMMAccessPort(ServiceRecord *svcRec)
{
	int		status = -1, s, p;
	AccessProtocols	*pAccessProtocols = NULL;

	if (getAccessProtocols(svcRec, &pAccessProtocols) == E_OK) {
		for (s = 0; s < (int)s_list_length(pAccessProtocols->pSequence); s++ ) {
			slist_t *protDescSeq = (slist_t*)s_list_nth_data(pAccessProtocols->pSequence, s);
			for (p = 0; p < (int)s_list_length(protDescSeq); p++ ) {
				ProtocolDescriptor *pPD = (ProtocolDescriptor*)s_list_nth_data(protDescSeq, p);
				if (UUID2value(&pPD->protocolId) == RFCOMM_UUID) {
					if (pPD->portNumber) {
						status = pPD->portNumber;
					}
					goto exit;
				}
			}
		}
	}
exit:
	return status;
}

int hasUUID(ServiceRecord *svcRec,  uint16_t uuid16)
{
	int		status = 0;
	slist_t		*searchPattern = NULL;
	UUID		uuid;

	makeUUIDFrom16Bits(&uuid, uuid16);
	s_list_append(&searchPattern, &uuid);
	if (!matchUUID(searchPattern, svcRec->targetPattern))
		status = -1;
	s_list_free(&searchPattern);
	return status;
}


