/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001,2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Original Author: Guruprasad Krishnamurthy <kgprasad@hotmail.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: sdpserver.c,v 1.32 2002/11/14 00:32:10 kds Exp $

   Contains the implementation of the SDP server itself including
   the main() of the server and it's initialization

   Fixes:
   		Imre Deak <ext-imre.deak@nokia.com>
		Dmitry Kasatkin
*/

#include <affix/config.h>

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <getopt.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <affix/bluetooth.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include <affix/sdpsrv.h>
#include "utils.h"
#include "servicedb.h"
#include "attr.h"
#include "cstate.h"
#include "request.h"

//#define SDP_THREADED

pthread_t	*thrIdArray;
pthread_attr_t	thrAttr;


fd_set		activeFdSet;
int		serverBTSockFd;
int		serverTCPSockFd;
int		maxFd;
int		argind;
static int	got_signal = 0;

/*
 ** Server and client socket addresses
 */
struct sockaddr_affix	serverBTAddr;
struct sockaddr_in	serverINAddr;

extern void startWorkers(int);

UUID sdpSrvUUID;
slist_t *versionNumberSeq = NULL;

ServiceRecord *sdpSvcRec;
ServiceRecord *publicBrowseGroupRoot;

/*
 ** List of version numbers supported
 ** by the SDP server. Add to this list
 ** when newer versions are supported
 */
const int sdpServerVnumEntries = 1;
SDPServerVersionNumber sdpVnumArray[1/*sdpServerVnumEntries*/] = {
	{1, 0} 
};

const char *pSvcName = "SDP Server";
const char *pProvName = "Nokia Research Center";
const char *pSvcDesc = "Affix Service Discovery server";

const char *pBrowseGroupName = "Public Browse Group Root";
const char *pGroupDesc = "Root of public group hierarchy";

static slist_t			*requestQueue_ = NULL;
static pthread_mutex_t		mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t		cond = PTHREAD_COND_INITIALIZER;

/*
** Setup the queue and initialize data structures
*/
int initQueue(void);
int addEntryToRequestQueue(SDPRequest *sdpRequest);
SDPRequest *removeFirstEntry(void);
int queueLength(void);
void flushQueue(void);


/*
 ** Setup the queue and initialize data structures
 */
int initQueue()
{
	return 0;
}


/*
 ** Add an entry to the queue. Grab the lock, append
 ** and signal on conditional to wake-up waiting threads
 */
int addEntryToRequestQueue(SDPRequest *sdpRequest)
{
	pthread_mutex_lock(&mutex);
	/*
	 ** Do stuff ..
	 */
	s_list_append(&requestQueue_, sdpRequest);
	pthread_cond_signal(&cond);
	pthread_mutex_unlock(&mutex);

	if (requestQueue_ != NULL)
		return 0;
	else
		return -1;
}

/*
 ** Take the first entry from the queue.
 ** Usually called by the "free" worker thread
 */
SDPRequest *removeFirstEntry()
{
	void * ptr;

	pthread_mutex_lock(&mutex);
	/*
	 ** Do stuff ..
	 */
	while ((ptr = s_list_nth_data(requestQueue_, 0)) == NULL) {
		pthread_cond_wait(&cond, &mutex);
	}
	s_list_remove(&requestQueue_, ptr);
	pthread_mutex_unlock(&mutex);

	return (SDPRequest *)ptr;
}

/*
 ** Find the number entries in the queue
 */
int queueLength()
{
	int length;

	pthread_mutex_lock(&mutex);
	/*
	 ** Do stuff ..
	 */
	length = s_list_length(requestQueue_);
	pthread_mutex_unlock(&mutex);

	return length;
}

/*
 ** Cleanup queue contents
 ** Called when server is about to shutdown
 */
void flushQueue()
{
}

/*
 ** Run method executed by the worker thread. Sleeps 
 ** until the request queue is populated. One of many worker(s)
 ** gets the newly added entry, and proceeds to process it
 */
void *run(void *arg)
{
	/*
	 **
	 */
	SDPRequest *ptr;

	ptr = NULL;
	for (;;) {
		ptr = removeFirstEntry();
		if (ptr != NULL) { 
			BTERROR("Non null entry : %d\n", (int)pthread_self());
			processRequest(ptr);
		}
	}

	return NULL;
}

/*
 ** Start "n" numbers of worker threads
 */
void startWorkers(int count)
{
	int i;

	pthread_attr_init(&thrAttr);
	//pthread_attr_setdetachstate(&thrAttr, PTHREAD_CREATE_DETACHED);
		
	thrIdArray = (pthread_t *)malloc(sizeof(pthread_t) * count);
	memset((void *)thrIdArray, 0, (sizeof(pthread_t) * count));

	BTDEBUG("Starting thread\n");
	for (i = 0; i < count; i++) {
		pthread_create(thrIdArray+i, &thrAttr, run, NULL);
		BTDEBUG("Thr Id : %ld\n", thrIdArray[i]);
	}
}


/*
 ** The service DB state is an attribute of the service record
 ** of the SDP server itself. This attribute is guaranteed to
 ** change if any of the contents of the service repository
 ** changes. This function updates the timestamp of value of
 ** the svcDBState attribute
 ** Set the SDP server DB. Simply a timestamp which is the marker
 ** when the DB was modified.
 */
void updateSDPServerDBTimestamp(void)
{
	uint32_t dbts = getTimeInSeconds();
	SDPData *pData = createSDPData(UnsignedInteger_32, (char *)&dbts);
	addToAttributeList(sdpSvcRec, AttrServiceDatabaseState, pData, 1);
	generateServiceRecordPduForm(sdpSvcRec);
}


int registerPublicBrowseGroupRoot(void)
{
	int		status = 0;
	char		*groupName;
	char		*groupDesc;
	char		*groupProvName;
	SDPData		*pSDPData, *attr;

	publicBrowseGroupRoot = createServiceRecord();
	publicBrowseGroupRoot->serviceRecordHandle = (ServiceRecordHandle)publicBrowseGroupRoot;
	addServiceRecord(publicBrowseGroupRoot);
	pSDPData = createSDPData(UnsignedInteger_32, &publicBrowseGroupRoot->serviceRecordHandle);
	addToAttributeList(publicBrowseGroupRoot, AttrServiceRecordHandle, pSDPData, 0);

	groupName = (char *)malloc(SequencePduFormSize);
	groupDesc = (char *)malloc(SequencePduFormSize);
	groupProvName = (char *)malloc(SequencePduFormSize);

	strcpy(groupName, pBrowseGroupName);
	strcpy(groupProvName, pProvName);
	strcpy(groupDesc, pGroupDesc);
	status = setInformationalAttributes(publicBrowseGroupRoot, groupName, groupProvName, groupDesc);

	attr = setServiceClassIDList(publicBrowseGroupRoot);
	status = addServiceClassID(attr, BrowseGroupDescServiceClassID);
	addUUID16ToTargetPattern(publicBrowseGroupRoot, BrowseGroupDescServiceClassID);
	
	status = setGroupID(publicBrowseGroupRoot, PublicBrowseGroup);
	addUUID16ToTargetPattern(publicBrowseGroupRoot, PublicBrowseGroup);

	generateServiceRecordPduForm(publicBrowseGroupRoot);

	return status;
}

/*
 ** The SDP server must present it's own service record to
 ** the service repository. This can be accessed by service
 ** discovery clients. This methods constructs a service record
 ** and stores it in the reposssitory
 */
int registerSDPServerService(void)
{
	int		status = 0, i;
	char		*svcName;
	char		*svcProvName;
	char		*svcDesc;
	SDPData		*pSDPData;
	SDPData		*pData;
	SDPData		*data;
	slist_t		*dataSeq = NULL;
	SDPData		*attr, *param;

	svcName = (char *)malloc(SequencePduFormSize);
	svcDesc = (char *)malloc(SequencePduFormSize);
	svcProvName = (char *)malloc(SequencePduFormSize);

	sdpSvcRec = createServiceRecord();
	/*
	 ** Force the svcRecHandle to be SDPServerServiceRecordHandle
	 */
	sdpSvcRec->serviceRecordHandle = SDPServerServiceRecordHandle;
	addServiceRecord(sdpSvcRec);
	pSDPData = createSDPData(UnsignedInteger_32, (char *)&sdpSvcRec->serviceRecordHandle);
	addToAttributeList(sdpSvcRec, AttrServiceRecordHandle, pSDPData, 0);

	/*
	 ** Add all attributes to service record. No need to
	 ** commit since we are the server and this record is
	 ** already in the serviceDB
	 */
	strcpy(svcName, pSvcName);
	strcpy(svcProvName, pProvName);
	strcpy(svcDesc, pSvcDesc);

	status = setInformationalAttributes(sdpSvcRec, svcName, svcProvName, svcDesc);

	attr = setServiceClassIDList(sdpSvcRec);
	status = addServiceClassID(attr, SDPServerServiceClassID);
	addUUID16ToTargetPattern(sdpSvcRec, SDPServerServiceClassID);

	/*
	 ** Set the version numbers supported, these are passed as arguments
	 ** to the server on command line. Now defaults to 1.0
	 ** Build the version number sequence first
	 */
	for (i = 0; i < sdpServerVnumEntries; i++) {
		uint16_t *pVNum;

		pVNum = (uint16_t *)malloc(sizeof(uint16_t));
		*pVNum = sdpVnumArray[i].major;
		*pVNum = (*pVNum << 8);
		*pVNum |= sdpVnumArray[i].minor;

		data = createSDPData(UnsignedInteger_16, pVNum);
		s_list_append(&dataSeq, data);
	}
	pData = createSDPData(SequenceSize8, dataSeq);
	addToAttributeList(sdpSvcRec, AttrVersionNumberList, pData, 0);

	status = setServiceID(sdpSvcRec, SDP_UUID);
	addUUID16ToTargetPattern(sdpSvcRec, SDP_UUID);

	attr = setProtocolDescriptorList(sdpSvcRec);
	param = addProtocolDescriptor(attr, L2CAP_UUID, SDP_PSM, 2, 1);

	attr = setBrowseGroupList(sdpSvcRec);
	status = addBrowseGroup(attr, PublicBrowseGroup);
	addUUID16ToTargetPattern(sdpSvcRec, PublicBrowseGroup);

	updateSDPServerDBTimestamp();

	return 0;
}


/*
 ** SDP server initialization on startup includes creating
 ** the BT PSM and TCP sockets over which service discovery clients
 ** and service registration clients respectively access the
 ** SDP server
 */
int initServer(void)
{
	int status = 0;

	status = sdp_init(SDP_SERVER);
	if (status != 0) {
		BTERROR("Could not find local BT device address, terminating");
		return -1;
	}
	initServiceDB();
	initContinuationStateCache();

	/*
	 ** Register the public browse group Root
	 */
	registerPublicBrowseGroupRoot();

	/*
	 ** Register the SDP server as a service
	 */
	registerSDPServerService();

	/*
	 * init locks to access Service Database
	 */
	initLock();

#if defined(SDP_THREADED)
	initQueue();
	/*
	 ** Start 1 worker thread; this should be
	 ** configurable later
	 */
	startWorkers(1);
#endif

	/*
	 ** Create our socket on PSM 1 and
	 ** set activeFdSet
	 */
	serverBTSockFd = socket(PF_AFFIX, SOCK_SEQPACKET, BTPROTO_L2CAP);
	if (serverBTSockFd < 0) {
		BTERROR("BT Socket creation failure:%s", strerror(errno));
		return serverBTSockFd;
	}

	serverBTAddr.family = AF_AFFIX;
	serverBTAddr.port = SDP_PSM;
	serverBTAddr.bda = BDADDR_ANY;
	status = bind(serverBTSockFd, (struct sockaddr *)&serverBTAddr, sizeof(serverBTAddr));
	if (status < 0) {
		BTERROR("BT bind failure:%s", strerror(errno));
		return status;
	}

	status = listen(serverBTSockFd, 5);
	if (status < 0)
		return status;

	FD_SET(serverBTSockFd, &activeFdSet);
	maxFd = serverBTSockFd;

	serverTCPSockFd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (serverTCPSockFd < 0) {
		BTERROR("TCP socket creation failure:%s", strerror(errno));
		return serverTCPSockFd;
	}
	serverINAddr.sin_family = AF_INET;
	serverINAddr.sin_port = SDP_TCP_PORT;
	status = bind(serverTCPSockFd, (struct sockaddr *)&serverINAddr, sizeof(serverINAddr));
	if (status < 0) {
		BTERROR("TCP bind failure:%s", strerror(errno));
		return status;
	}

	status = listen(serverTCPSockFd, 5);
	FD_SET(serverTCPSockFd, &activeFdSet);
	maxFd = serverTCPSockFd;

	return status;
}

void exitServer(void)
{
	resetServiceDB();
	cleanupContinuationStateBuffer();
	close(serverBTSockFd);
	close(serverTCPSockFd);
}


/*
 ** This is the SDP server main thread
 ** After appropriate initialization, uses the select()
 ** to field inputs from both the BT and TCP sockets.
 **
 ** Calls the processRequest() if single threaded else
 ** extracts the request and populates the request queue 
 ** for servicing by one or more worker threads that are
 ** active
 */

void sdpServerSignalHandler(int sig)
{
	got_signal = sig;
}

void usage(void)
{
	fprintf(stderr, "btsdp %s [-d] [-b]\n"
			"  -d: start as daemon\n  -b: with debug info\n", affix_version);
}

int main(int argc, char **argv)
{
	/*
	 ** Create our socket for connection
	 ** requests. Start worker threads
	 */
	int	status = 0, lind, c;
	int	btdaemon = 0;
	int	clientCount = 0;

	struct option	opts[] = {
		{"help", 0, 0, 'h'},
		{"daemon", 0, 0, 'd'},
		{"debug", 0, 0, 'b'},
		{0, 0, 0, 0}
	};

	openlog(NULL/*argv[0]*/, 0, LOG_DAEMON);

	btlog_mask = BTDEBUG_MODULE;
	for (;;) {
		c = getopt_long(argc, argv, "+hdbD", opts, &lind);
		if (c == -1)
			break;
		switch (c) {
			case 'h':
				usage();
				return -1;
				break;
			case 'd':
				btdaemon = 1;
				break;
			case 'D':
				btdaemon = 2;	// do not fork
				break;
			case 'b':
				btlog_mask = BTDEBUG_MODULE;
				break;
			case ':':
				printf("Missing parameters for option: %c\n", optopt);
				return 1;
				break;
			case '?':
				printf("Unknown option: %c\n", optopt);
				return 1;
				break;
		}
	}

	argind = optind;

	BTINFO("%s %s started.", argv[0], affix_version);

	status = initServer();
	if (status != 0) {
		BTERROR("Terminating, init error");
		return -1;
	}

	if (btdaemon && become_daemon(btdaemon != 2))
		return 0;

	signal(SIGINT, sdpServerSignalHandler);
	signal(SIGTERM, sdpServerSignalHandler);
	signal(SIGABRT, sdpServerSignalHandler);
	signal(SIGQUIT, sdpServerSignalHandler);

	for (;;) {
		int		numFds;
		int		newFd;
		fd_set		checkFds;
		struct timeval	tv;
		int 		fd;

		FD_ZERO(&checkFds);
		checkFds = activeFdSet;
		if (!clientCount)
			tv.tv_sec = 3;
		else
			tv.tv_sec = 1000;
		tv.tv_usec = 0;
		
		/*
		 ** Check on active fds
		 */
		BTDEBUG("selecting new FD to read, clientCount: %d\n", clientCount);
		numFds = select(maxFd+1, &checkFds, NULL, NULL, &tv);
		BTDEBUG("Selected: %d, %s (%d)", numFds, strerror(errno), errno);
		if (numFds < 0) {
			BTDEBUG("Select error:%s", strerror(errno));
			goto exit;
		} else if (numFds == 0) {
			/* timeout happend */
			if (clientCount == 0) {
				BTDEBUG("No connected clients, quit...\n");
				goto exit;
			}
		}

		if (FD_ISSET(serverBTSockFd, &checkFds)) {
			/*
			 ** A new Bluetooth connection ..
			 */
			struct sockaddr_affix	clientBTAddr;
			socklen_t		sockLength = sizeof(clientBTAddr);

			numFds--;
			newFd = accept(serverBTSockFd, (struct sockaddr *)&clientBTAddr, &sockLength);
			BTDEBUG("new bt fd : %d", newFd); 
			if (newFd != -1) {
				if (newFd > maxFd)
					maxFd = newFd;
				FD_SET(newFd, &activeFdSet);
				clientCount++;
			} else
				BTERROR("Error accepting Bluetooth client connection:%s", strerror(errno));
				
		} else if (FD_ISSET(serverTCPSockFd, &checkFds)) {
			/*
			 ** A new TCP connection
			 */
			struct sockaddr_in	clientINAddr;
			socklen_t		sockLength = sizeof(clientINAddr);

			numFds--;
			newFd = accept(serverTCPSockFd, (struct sockaddr *)&clientINAddr, &sockLength);
			BTDEBUG("new tcp fd : %d", newFd);
			if (newFd != -1) {
				if (newFd > maxFd)
					maxFd = newFd;
				FD_SET(newFd, &activeFdSet);
				clientCount++;
			} else
				BTERROR("Error accepting TCP client connection:%s", strerror(errno));
		} else {
			for (fd = 0; fd <= maxFd && numFds; fd++) {
				SDPRequest	*sdpReq;
				char		*realBuf;
				int		bufSize;

				if (fd == serverBTSockFd || fd == serverTCPSockFd)
					continue;
				if (FD_ISSET(fd, &checkFds)) {
					int bytesRead = 0;
					SdpPduHeader sdpHeader;

					numFds--;
					BTDEBUG("Reading from : %d", fd);
					bytesRead = recv(fd, (void *)&sdpHeader, sizeof(SdpPduHeader), MSG_PEEK);
					BTDEBUG("Bytes read : %d", bytesRead);
					if ((bytesRead < 0) || (bytesRead == 0)) {
						if (bytesRead < 0) {
							BTERROR("Read error:%s", strerror(errno));
						}
						handleServiceDisconnectRequest(fd);
						closeConnection(fd);
						BTDEBUG("Conn closed: %d, read: %d", fd, bytesRead);
						clientCount--;
						FD_CLR(fd, &activeFdSet);
						if (fd == maxFd)
							maxFd--;
					} else {
						bufSize = sizeof(SdpPduHeader) + ntohs(sdpHeader.paramLength);
						BTDEBUG("Allocating size : %d", bufSize);
						realBuf = (char *)malloc(bufSize);
						bytesRead = recv(fd, realBuf, bufSize, 0);
						BTDEBUG("Bytes read : %d", bytesRead);

						if ((bytesRead < 0) || (bytesRead == 0)) {
							if (bytesRead < 0) {
								BTERROR("Read error:%s", strerror(errno));
							}
							closeConnection(fd);
							BTDEBUG("Conn closed : %d", fd);
							clientCount--;
							FD_CLR(fd, &activeFdSet);
							if (fd == maxFd)
								maxFd--;
						} else {
							BTDEBUG("Bytes read : %d", bytesRead);
							sdpReq = (SDPRequest*)malloc(sizeof(SDPRequest));
							sdpReq->requestSocket = fd;
							sdpReq->requestBuffer = realBuf;
						#if defined(SDP_THREADED)
							addEntryToRequestQueue(sdpReq);
						#else
							BTDEBUG("Processing request");
							processRequest(sdpReq);
							BTDEBUG("Out of process request");
						#endif
						}
					}
				}
			}
		}
	}
exit:
	exitServer();
	return status;
}


