/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001,2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Original Author: Guruprasad Krishnamurthy <kgprasad@hotmail.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: sdpsrv.c,v 1.16 2002/11/22 17:16:27 kds Exp $

   Contains implementation of SDP service registration APIs
 
   Contains utlities for use by SDP service registration clients
   to add attributes to a service record.

   Fixes:
		Manel Guerrero			: 
		Dmitry Kasatkin			: SDPServerHandle changed to complex type
		Dmitry Kasatkin			: cleanup
*/

#include <affix/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <affix/sdp.h>
#include <affix/sdpsrv.h>
#include <affix/sdpclt.h>
#include "utils.h"
#include "attr.h"


static ServiceAttributeIdentifier svcRecHandleAttrId = AttrServiceRecordHandle;

/*
 ** Returns the internal state of the service record
 ** pointed to by svcRecHandle
 */
ServiceRecordInternalState getServiceRecordInternalState(ServiceRecord *svcRec)
{
	if (svcRec != NULL) {
		return svcRec->internalState;
	} else
		return doesNotExist;
}

/*
 **
 ** Registers a newly created ServiceRecord with the SDP server
 ** specified by the SDPServerHandle.
 **
 ** It is incorrect to call this method on a ServiceRecord that
 ** has been already registered with the server.
 **
 ** Returns a non-null value (a pointer) to a service
 ** record if successful, else -1
 */
int registerServiceRecord(SDPServerHandle srvHandle, ServiceRecord *svcRec)
{
	char	*pdata;
	int	status = 0;
	char	*requestBuffer;
	char	*responseBuffer;
	int	requestSize;
	int	responseSize;

	SdpPduHeader	*pduRequestHeader;
	SdpPduHeader	*pduResponseHeader;
	SDPData		*pData;

	ServiceRecordHandle	svcRecHandle;


	if (svcRec->internalState != createdRegisterReady)
		return E_FAILURE;
	
	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceRegistrationRequest;
	pduRequestHeader->transactionId = htons(generateTransactionId());

	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);                       

	generateServiceRecordPduForm(svcRec);
	memcpy(pdata, svcRec->pduForm.data, svcRec->pduForm.dataSize);
	requestSize += svcRec->pduForm.dataSize;

	pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));

	status = sendRequestWaitForResponse(srvHandle, requestBuffer,
			responseBuffer, requestSize, &responseSize);

	if (status == E_OK) {
		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		pdata = responseBuffer + sizeof(SdpPduHeader);                 
		if ((pduResponseHeader->pduId == SDP_ServiceRegistrationResponse) &&
				(status == E_OK)) {
			svcRecHandle = ntohl(__get_u32(pdata));
			svcRec->serviceRecordHandle = svcRecHandle;
			pData = createSDPData(UnsignedInteger_32, 
					(char *)&svcRecHandle);
			addToAttributeList(svcRec, svcRecHandleAttrId, pData, 1);
			svcRec->internalState = registered;
		} else
			status = -1;
	}
	free(requestBuffer);
	free(responseBuffer);
	return status;
}

/*
 ** Deletes a service record pointed to by
 ** svcRecHandle. Note that you still need
 ** to call commitServiceRecord() to effect
 ** an actual delete in the SDP server DB
 **
 ** Returns 0 if successful or -1 if there
 ** is a failure.
 **
 */
int deleteServiceRecord(SDPServerHandle srvHandle, ServiceRecord *svcRec)
{
	char	*pdata;
	int	status = 0;
	char	*requestBuffer;
	char	*responseBuffer;
	int	requestSize = 0;
	int	responseSize = 0;

	SdpPduHeader	*pduRequestHeader;
	SdpPduHeader	*pduResponseHeader;

	ServiceRecordHandle	svcRecHandle = 0;

	svcRecHandle = svcRec->serviceRecordHandle;
	if ((svcRecHandle == SDPServerServiceRecordHandle) ||
			((svcRec->internalState != registered) &&
			 (svcRec->internalState != created) && /* FIXME POPO TAROM */
			 (svcRec->internalState != modifiedUpdateReady)))
		return E_INVALID_ARGS;

	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceRemovalRequest;
	pduRequestHeader->transactionId = htons(generateTransactionId());

	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);                       
	__put_u32(pdata, htonl(svcRecHandle));
	requestSize += sizeof(ServiceRecordHandle);

	pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));
	status = sendRequestWaitForResponse(srvHandle, requestBuffer,
			responseBuffer, requestSize, &responseSize);
	if (status == E_OK) {
		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		pdata = responseBuffer + sizeof(SdpPduHeader);                 
		status = __get_u16(pdata);
		if ((pduResponseHeader->pduId == SDP_ServiceRemovalResponse) &&
				(status == E_OK)) {
			BTDEBUG("Removing local copy\n");
			freeServiceRecord(svcRec);
			svcRec->internalState = doesNotExist;
		} else
			status = -1;
	}
	free(requestBuffer);
	free(responseBuffer);
	return status;
}

int __deleteServiceRecord(ServiceRecord *svcRec)
{
	SDPServerHandle srvHandle;
	int		status;


	srvHandle = SDPOpenLocalServerConnection();
	if (srvHandle < 0)
		return srvHandle;

	status = deleteServiceRecord(srvHandle, svcRec);

	SDPCloseServerConnection(srvHandle);
	return status;
}


int deleteServiceRecordByHandle(SDPServerHandle srvHandle, ServiceRecordHandle svcRecHandle)
{
	char	*pdata;
	int	status = 0;
	char	*requestBuffer;
	char	*responseBuffer;
	int	requestSize = 0;
	int	responseSize = 0;

	SdpPduHeader	*pduRequestHeader;
	SdpPduHeader	*pduResponseHeader;


	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceRemovalRequest;
	pduRequestHeader->transactionId = htons(generateTransactionId());

	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);                       
	__put_u32(pdata, htonl(svcRecHandle));
	requestSize += sizeof(ServiceRecordHandle);

	pduRequestHeader->paramLength = 
		htons(requestSize - sizeof(SdpPduHeader));
	status = sendRequestWaitForResponse(srvHandle, requestBuffer,
			responseBuffer, requestSize, &responseSize);
	if (status == E_OK) {
		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		pdata = responseBuffer + sizeof(SdpPduHeader);                 
		status = __get_u16(pdata);
		if ((pduResponseHeader->pduId == SDP_ServiceRemovalResponse) &&
				(status == E_OK)) {
			BTDEBUG("Removing local copy\n");
		} else
			status = -1;
	}
	free(requestBuffer);
	free(responseBuffer);
	return status;
}

int __deleteServiceRecordByHandle(ServiceRecordHandle svcRecHandle)
{
	SDPServerHandle srvHandle;
	int		status;

	srvHandle = SDPOpenLocalServerConnection();
	if (srvHandle < 0)
		return srvHandle;

	status = deleteServiceRecordByHandle(srvHandle, svcRecHandle);

	SDPCloseServerConnection(srvHandle);
	return status;
}



/*
 ** Use this function to modify an already created
 ** service record
 */
int updateServiceRecord(SDPServerHandle srvHandle, ServiceRecord *svcRec)
{
	char *pdata;
	int status = 0;
	char *requestBuffer;
	char *responseBuffer;
	int requestSize;
	int responseSize;
	SdpPduHeader *pduRequestHeader;
	SdpPduHeader *pduResponseHeader;
	ServiceRecordHandle svcRecHandle;

	svcRecHandle = svcRec->serviceRecordHandle;
	if ((svcRecHandle == SDPServerServiceRecordHandle) ||
			((svcRec->internalState != modifiedUpdateReady)))
		return E_INVALID_ARGS;

	requestBuffer = (char *)malloc(RequestBufferSize);
	if (!requestBuffer)
		return E_FAILURE;
	responseBuffer = (char *)malloc(ResponseBufferSize);
	if (!responseBuffer) {
		free(requestBuffer);
		return E_FAILURE;
	}
	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduRequestHeader->pduId = SDP_ServiceUpdateRequest;
	pduRequestHeader->transactionId = htons(generateTransactionId());

	pdata = (char *)(requestBuffer + sizeof(SdpPduHeader));
	requestSize = sizeof(SdpPduHeader);                       

	__put_u32(pdata, htonl(svcRecHandle));
	requestSize += sizeof(ServiceRecordHandle);
	pdata += sizeof(ServiceRecordHandle);

	generateServiceRecordPduForm(svcRec);
	memcpy(pdata, svcRec->pduForm.data, svcRec->pduForm.dataSize);
	requestSize += svcRec->pduForm.dataSize;

	pduRequestHeader->paramLength = htons(requestSize - sizeof(SdpPduHeader));
	status = sendRequestWaitForResponse(srvHandle, requestBuffer,
			responseBuffer, requestSize, &responseSize);
	BTDEBUG("Send req status : %d\n", status);
	if (status == E_OK) {
		pduResponseHeader = (SdpPduHeader *)responseBuffer;
		pdata = responseBuffer + sizeof(SdpPduHeader);                 
		status = __get_u16(pdata);
		if ((pduResponseHeader->pduId == SDP_ServiceUpdateResponse) &&
				(status == E_OK)) {
			svcRec->internalState = registered;
		} else
			status = -1;
	}
	free(requestBuffer);
	free(responseBuffer);
	return status;
}

int __updateServiceRecord(ServiceRecord *svcRec)
{
	SDPServerHandle srvHandle;
	int		status;


	srvHandle = SDPOpenLocalServerConnection();
	if (srvHandle < 0)
		return srvHandle;

	status = updateServiceRecord(srvHandle, svcRec);

	SDPCloseServerConnection(srvHandle);
	return status;
}

/* ------------------------------------------------------------------------ */

/*
 ** NOTE that none of the setXXX() functions below will
 ** actually update the SDP server, unless the
 ** {register, update}ServiceRecord() function is invoked.
 **
 **
 */


void setInternalState(ServiceRecord *svcRec)
{
	BTDEBUG("Internal state of svcRec : %d", svcRec->internalState);
	if (svcRec->internalState == created) {
		if ((getSDPData(svcRec, AttrServiceClassIDList) != NULL) &&
				(getSDPData(svcRec, AttrProtocolDescriptorList) != NULL)) {
			svcRec->internalState = createdRegisterReady;
			BTDEBUG("Internal state of svcRec reg ready : %d", svcRec->internalState);
		} else if (getSDPData(svcRec, AttrGroupID) != NULL) {
			/*
			 ** This could be a browse group descriptor, which
			 ** need not contain mandatory attributes
			 */
			svcRec->internalState = createdRegisterReady;
			BTDEBUG("Internal state of svcRec reg ready : %d", svcRec->internalState);
		}
	} else if (svcRec->internalState == registered) {
		svcRec->internalState = modifiedUpdateReady;
	}
}

int addBasicAttribute(ServiceRecord *svcRec, DataTypeDescriptor dtd, void *pValue, 
		ServiceAttributeIdentifier attributeId)
{
	SDPData *pSDPData;
	
	if ((svcRec == NULL) || (pValue == NULL))
		return -1;

	pSDPData = createSDPData(dtd, (char *)pValue);
	if (pSDPData == NULL)
		return -1;

	addToAttributeList(svcRec, attributeId, pSDPData, 1);
	setInternalState(svcRec);

	return 0;
}


/*
 ** Set the information attributes of the service
 ** pointed to by svcRecHandle. The set of attributes
 ** include service name, description and provider name
 **
 ** Returns 0 if successful and -1 if there is
 ** a failure.
 */
int setInformationalAttributes(ServiceRecord *svcRec, char *serviceName, 
		char *providerName, char *serviceDescription)
{
	if ((svcRec == NULL) || (svcRec->internalState == doesNotExist))
		return -1;

	addBasicAttribute(svcRec, TextStringSize8, serviceName, AttrServiceName_Primary);
	addBasicAttribute(svcRec, TextStringSize8, providerName, AttrServiceProviderName_Primary);
	addBasicAttribute(svcRec, TextStringSize8, serviceDescription, AttrServiceDescription_Primary);
	setInternalState(svcRec);

	return 0;
}


/*
 ** This function sets the serviceClassID of the
 ** service pointed to by ServiceRecord,
 ** to the list specified by svcClassIDList.
 ** Note that the identifiers need to be in
 ** sorted order from the most specific to the
 ** most generic service class that this service
 ** conforms to.
 **
 ** Using a value of NULL for svcClassIDList has
 ** effect of removing this attribute (if previously set)
 ** 
 ** This function replaces the existing ClassID
 ** list with a new one specified by svcClassIDList
 ** if one is present.
 **
 ** Returns 0 if successful or
 ** -1 if there is a failure.
 **
 */
SDPData *setServiceClassIDList(ServiceRecord *svc)
{
	SDPData	*attr;
	int	err;

	attr = createSDPData(SequenceSize8,  NULL);
	if (attr == NULL)
		return NULL;

	err = addToAttributeList(svc, AttrServiceClassIDList, attr, 0);
	if (err) {
		freeSDPData(attr);
		return NULL;
	}
	return attr;
}

int addServiceClassID(SDPData *attr, UUID16Bit uuid)
{
	SDPData		*data;

	data = createSDPData(UUID_16, &uuid);
	if (data == NULL)
		goto err;
	addSDPDataSeq(attr, data);
	
	return 0;
err:
	freeSDPData(data);
	return -1;
}


/*
 ** This function sets the BrowseGroupList of the
 ** service pointed to by ServiceRecord,
 ** to the list specified by browseGroupSeq.
 **
 ** Using a value of NULL for browseGroupSeq has
 ** effect of removing this attribute (if previously set)
 ** 
 ** Returns 0 if successful or
 ** -1 if there is a failure.
 **
 */
SDPData *setBrowseGroupList(ServiceRecord *svc)
{
	SDPData	*attr;
	int	err;

	attr = createSDPData(SequenceSize8,  NULL);
	if (attr == NULL)
		return NULL;

	err = addToAttributeList(svc, AttrBrowseGroupList, attr, 0);
	if (err) {
		freeSDPData(attr);
		return NULL;
	}

	return attr;
}

int addBrowseGroup(SDPData *attr, UUID16Bit uuid)
{
	SDPData		*data;

	data = createSDPData(UUID_16, &uuid);
	if (data == NULL)
		goto err;
	addSDPDataSeq(attr, data);

	return 0;
err:
	freeSDPData(data);
	return -1;
}


/*
 ** This function sets the access protocols
 ** of the service specified by ServiceRecord.
 ** to the value specified in "accessProtocols"
 **
 ** Note that if there are alternate mechanisms by 
 ** which the service is accessed, then they should 
 ** be specified as sequences 
 **
 ** Using a value of NULL for acessProtocols has
 ** effect of removing this attribute (if previously set)
 ** 
 ** This function replaces the exisiting AccessProtocols
 ** structure (if any) with the new one specified.
 **
 ** Returns 0 if successful or
 ** -1 if there is a failure.
 **
 */

SDPData *setProtocolDescriptorList(ServiceRecord *svc)
{
	SDPData	*attr;
	int	err;

	attr = createSDPData(SequenceSize8,  NULL);
	if (attr == NULL)
		return NULL;

	err = addToAttributeList(svc, AttrProtocolDescriptorList, attr, 0);
	if (err) {
		freeSDPData(attr);
		return NULL;
	}
	
	return attr;
}

SDPData *addProtocolDescriptorList(SDPData *attr)
{
	SDPData		*data;

	data = createSDPData(SequenceSize8,  NULL);
	if (data == NULL)
		return NULL;

	/* change type to alternative */
	attr->dtd = AlternativeSize8;
	addSDPDataSeq(attr, data);

	return data;
}


SDPData *addProtocolDescriptor(SDPData *attr, UUID16Bit uuid, 
		PortNumber portNumber, int portSize, VersionNumber version)
{
	SDPData		*data;
	slist_t		*dataSeq = NULL;

	data = createSDPData(UUID_16, &uuid);
	if (data == NULL)
		goto err;
	s_list_append(&dataSeq, data);

	if (portNumber != 0xFFFF) {
		if (portSize == 1) {
			uint8_t	port = portNumber;
			data = createSDPData(UnsignedInteger_8, &port);
		} else
			data = createSDPData(UnsignedInteger_16, &portNumber);
		if (data == NULL)
			goto err;
		s_list_append(&dataSeq, data);
	}

	if (version != 0xffff) {
		data = createSDPData(UnsignedInteger_16, &version);
		if (data == NULL)
			goto err;
		s_list_append(&dataSeq, data);
	}

	/* create root element */
	data = createSDPData(SequenceSize8, dataSeq);
	if (data == NULL)
		goto err;
	
	addSDPDataSeq(attr, data);

	return data;
err:
	freeSDPDataSeq(dataSeq);
	return NULL;
}


/*
 ** This function sets the "LanguageBase"
 ** attributes of the service record pointed
 ** to by svcRecHandle to the value specified 
 ** in "langAttrList".
 **
 ** "langAttrList" is a linked list of "LanguageAttributes"
 ** objects, one for each language in which user visible
 ** attributes are present in the service record.
 **
 ** Using a value of NULL for langAttrList has
 ** effect of removing this attribute (if previously set)
 ** 
 ** This function replaces the exisiting LanguageAttributes
 ** structure (if any) with the new one specified.
 **
 ** Returns 0 if successful or
 ** -1 if there is a failure.
 **
 */
SDPData *setLanguageBaseAttributeIDList(ServiceRecord *svcRec)
{
	SDPData	*attr;
	int	err;

	attr = createSDPData(SequenceSize8,  NULL);
	if (attr == NULL)
		return NULL;

	err = addToAttributeList(svcRec, AttrLanguageBasedAttrIDList, attr, 0);
	if (err) {
		freeSDPData(attr);
		return NULL;
	}
	
	return attr;
}

int addLanguageBase(SDPData *attr, uint16_t lang, uint16_t encoding, uint16_t offset)
{
	SDPData		*data;

	data = createSDPData(UnsignedInteger_16, &lang);
	if (data == NULL)
		goto err;
	addSDPDataSeq(attr, data);

	data = createSDPData(UnsignedInteger_16, &encoding);
	if (data == NULL)
		goto err;
	addSDPDataSeq(attr, data);

	data = createSDPData(UnsignedInteger_16, &offset);
	if (data == NULL)
		goto err;
	addSDPDataSeq(attr, data);
	
	return 0;
err:
	freeSDPData(data);
	return -1;
}


/*
 ** This function sets the "ServiceInfoTimeToLive"
 ** attribute of the service. This is the number
 ** of seconds that this service record is guaranteed
 ** not to change after the record was obtained by the
 ** client from the SDP server.
 **
 ** It is not clear if this attribute
 ** is set by the SDP server or the owner or the
 ** service record. Just in case ..
 **
 ** Returns 0 if successful or
 ** -1 if there is a failure.
 **
 */
int setServiceInfoTimeToLive(ServiceRecord *svcRec, ServiceInfoTimeToLive svcTTL)
{
	int status = 0;

	addBasicAttribute(svcRec, UnsignedInteger_32, &svcTTL, AttrServiceInfoTimeToLive);

	return status;
}


/*
 ** This function sets the "ServiceRecordState"
 ** attribute of the service. This attribute is
 ** guaranteed to change if there is any kind of
 ** modification to the service record. 
 ** 
 ** Again it is not clear if this attribute is set
 ** by the SDP server or the service provider. But
 ** this one seems more likely an attribute that
 ** is set by the service provider.
 **
 ** Returns 0 if successful or
 ** -1 if there is a failure.
 **
 */
int setServiceRecordState(ServiceRecord *svcRec, ServiceRecordState svcRecState)
{
	int status = 0;

	addBasicAttribute(svcRec, UnsignedInteger_32, &svcRecState, AttrServiceRecordState);
	return status;
}


/*
 ** This function sets the "ServiceID"
 ** attribute of the service. 
 ** 
 ** This is the UUID of the service. 
 ** 
 ** Returns 0 if successful or
 ** -1 if there is a failure.
 **
 */
int setServiceID(ServiceRecord *svcRec, UUID16Bit svcUUID)
{
	int status = 0;

	addBasicAttribute(svcRec, UUID_16, &svcUUID, AttrServiceID);

	return status;
}


/*
 ** This function sets the "GroupID"
 ** attribute of the service record
 ** defining a group. 
 ** 
 ** This is the UUID of the group. 
 ** 
 ** Returns 0 if successful or
 ** -1 if there is a failure.
 **
 */
int setGroupID(ServiceRecord *svcRec, UUID16Bit groupUUID)
{
	int status = 0;

	addBasicAttribute(svcRec, UUID_16, &groupUUID, AttrGroupID);

	return status;
}


/*
 ** This function sets the "ServiceAvailability"
 ** attribute of the service pointed to svcRecHandle
 ** to the value specified in svcAvail.
 **
 ** Note that this represents the relative availability
 ** of the service 0x00 means compeletely unavailable to
 ** 0xFF meaning maximum availability.
 ** 
 ** Returns 0 if successful or
 ** -1 if there is a failure.
 **
 */
int setServiceAvailability(ServiceRecord *svcRec, ServiceAvailability svcAvail)
{
	int status = 0;

	addBasicAttribute(svcRec, UnsignedInteger_8, &svcAvail, AttrServiceAvailability);

	return status;
}


/*
 ** This function sets the "ProfileDescriptorList"
 ** attribute of the service record pointed
 ** to by svcRecHandle to the value specified 
 ** in "profileDesc".
 **
 ** Each element in the slist_t is an object of type
 ** ProfileDescriptor which is a definition of the
 ** Bluetooth profile that this service conforms to.
 **
 ** Using a value of NULL for profileDesc has
 ** effect of removing this attribute (if previously set)
 ** 
 ** This function replaces the exisiting BTProfileDescritorList
 ** structure (if any) with the new one specified.
 **
 ** Returns 0 if successful or
 ** -1 if there is a failure.
 **
 */

SDPData *setProfileDescriptorList(ServiceRecord *svc)
{
	SDPData	*attr;
	int	err;

	attr = createSDPData(SequenceSize8,  NULL);
	if (attr == NULL)
		return NULL;

	err = addToAttributeList(svc, AttrProfileDescriptorList, attr, 0);
	if (err) {
		freeSDPData(attr);
		return NULL;
	}
	
	return attr;
}

int addProfileDescriptor(SDPData *attr, UUID16Bit uuid, VersionNumber version)
{
	SDPData		*data;
	slist_t		*dataSeq = NULL;

	data = createSDPData(UUID_16, &uuid);
	if (data == NULL)
		goto err;
	s_list_append(&dataSeq, data);

	data = createSDPData(UnsignedInteger_16, &version);
	if (data == NULL)
		goto err;
	s_list_append(&dataSeq, data);

	data = createSDPData(SequenceSize8, dataSeq);
	if (data == NULL)
		goto err;
	
	addSDPDataSeq(attr, data);

	return 0;
err:
	freeSDPDataSeq(dataSeq);
	return -1;
}

/*
 ** Sets various URL attributes of the service
 ** pointed to by svcRecHandle. The URL include
 **
 ** ClientExecutableURL : a URL to the client's
 **   platform specific (WinCE, PalmOS) executable
 **   code that can be used to access this service.
 **
 ** DocumentationURL : a URL pointing to service documentation
 **
 ** IconURL : a URL to an icon that can be used to represent
 **   this service.
 **
 ** Note that you need to pass NULL for any URLs
 ** that you don't want to set or remove
 **
 ** Returns 0 if successful and -1 if there is
 ** a failure.
 **
 */
int setURLAttributes(ServiceRecord *svcRec, char *clientExecURL, char *docURL, char *iconURL)
{
	int status = 0;
	DataTypeDescriptor dtd;

	dtd = URLStringSize8;

	if ((svcRec != NULL) && (svcRec->internalState != doesNotExist)) {
		addBasicAttribute(svcRec, dtd, clientExecURL, AttrClientExecutableURL);
		addBasicAttribute(svcRec, dtd, docURL, AttrDocumentationURL);
		addBasicAttribute(svcRec, dtd, iconURL, AttrIconURL);
		setInternalState(svcRec);
	}
	else 
		status = -1;

	return status;
}

/* -------------------------------- */

ServiceRecord *createRFCOMMServiceRecord(int port, int flags)
{
	ServiceRecord	*svcRec;
	SDPData		*attr, *param;

	svcRec = createServiceRecord();
	if (svcRec == NULL) {
		BTERROR("createServiceRecord failed");
		return NULL;
	}

	attr = setBrowseGroupList(svcRec);
	if (attr == NULL) {
		BTERROR("setBrowseGroupList failed");
		freeServiceRecord(svcRec);
		return NULL;
	}
	addBrowseGroup(attr, PublicBrowseGroup);

	attr = setProtocolDescriptorList(svcRec);
	if (attr == NULL) {
		BTERROR("setProtocolDescriptorList failed");
		freeServiceRecord(svcRec);
		return NULL;
	}
	
	param = addProtocolDescriptor(attr, L2CAP_UUID, 0xffff, 0, 0xffff);
	param = addProtocolDescriptor(attr, RFCOMM_UUID, port, 1, 0xffff);
	if (flags) {
		param = addProtocolDescriptor(attr, OBEX_UUID, 0xffff, 0, 0xffff);
	}
		
	return svcRec;
}


