/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001,2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Original Author: Guruprasad Krishnamurthy <kgprasad@hotmail.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
   $Id: utils.c,v 1.32 2002/11/15 16:47:06 kds Exp $

   SDP client utilities to perform GIAC, connect to an SDP server,
   and to initialize the SDP machinery

   Fixes:
	    Dmitry Kasatkin		: SDPServerConnection changed to complex type
*/

#include <affix/config.h>

#include <stdio.h>

#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include <affix/sdpsrv.h>
#include "utils.h"
#include "cstate.h"
#include "attr.h"


static int	sdp_up = 0;
static int	sdp_flags = 0;
static pid_t	sdp_pid = 0;
OperationMode	sdp_mode;

/*
 ** SDP Initialization procedure. Sets up SDP internal
 ** data structures. Must be called before any other
 ** SDP function can be invoked.
 **
 ** INPUT :
 **
 **   NONE
 **
 ** OUTPUT :
 ** 
 **   E_OK : Initialization success
 **
 **   E_FAILURE : Initialization failure
 **
 */
int SDPInit(int flags)
{
	int	err = 0;
	
	if (!sdp_up) {
		err = sdp_init(SDP_CLIENT);
		if (!err) {
			sdp_up = 1;
			sdp_flags = flags;
		}
	}
	return err;
}

void SDPCleanup(void)
{
	if (sdp_up) {
		sdp_shutdown();
		sdp_up = 0;
		sdp_flags = 0;
	}
}
	
/*
 ** Create a connection (L2CAP) to the specified
 ** Bluetooth device (BD_ADDR). 
 **
 ** Note : The identification of the device on which
 **        the SDP server is known to run is outside 
 **        the scope of this function. For a utility
 **        function to do this, see sdp_utils.h
 **
 ** INPUT :
 **  
 **  char * bdAddr :
 **    Address of the device where the SDP server is known
 **    to run. In the format xx:yy:zz:ll:mm
 **
 **  SDPServerHandle :
 **    Contains a handle to the SDP server on the remote
 **    device. All subsequent transactions must be carried
 **    out using this
 **
 */
SDPServerHandle SDPOpenServerConnection(struct sockaddr_affix *saddr)
{
	int		status = 0;
	int		clientSockFd;
	uint16_t	port;

	if (saddr->family == PF_INET) {
		/*
		 ** Code for a local SDP server connection
		 */
		struct sockaddr_in	*inAddr = (struct sockaddr_in*)saddr;
		BTDEBUG("Attempt to open a TCP socket, local client\n");
		clientSockFd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (clientSockFd < 0)
			return clientSockFd;
		port = inAddr->sin_port;
		inAddr->sin_port = SDP_TCP_PORT;
		BTDEBUG("TCP fd : %d\n", clientSockFd);
		BTDEBUG("Local addr : %x\n", inAddr->sin_addr.s_addr);
		status = connect(clientSockFd, (struct sockaddr *)inAddr, sizeof(*inAddr));
		if (status < 0) {
			close(clientSockFd);
			BTERROR("TCP connect error:%s", strerror(errno));
		} else {
			BTDEBUG("Connected\n");
		}
		inAddr->sin_port = port;
	} else if (saddr->family == PF_AFFIX) {                                             
		/*
		 ** Create a BT socket for ourselves
		 */
		BTDEBUG("Attempt to open BT connection\n");
		BTDEBUG("BT addr str : %s\n", bda2str(&saddr->bda));
		clientSockFd = socket(PF_AFFIX, SOCK_SEQPACKET, BTPROTO_L2CAP);
		if (clientSockFd < 0)
			return clientSockFd;
		port = saddr->port;
		saddr->port = SDP_PSM;
		saddr->local = BDADDR_ANY;
		BTDEBUG("Connecting\n");
		status = connect(clientSockFd, (struct sockaddr*)saddr, sizeof(*saddr));          
		if (status < 0) {
			close(clientSockFd);
			BTDEBUG("connect error:%s", strerror(errno));
			
		} else {
			BTDEBUG("Connect success, mtu: %d\n", l2cap_getmtu(clientSockFd));
		}
		//printf("Connect success, mtu: %d\n", l2cap_getmtu(clientSockFd));
		saddr->port = port;
	} else
		return -1;

	if (status == 0)
		return clientSockFd;
	else {
		switch (errno) {
			case ETIMEDOUT:
				return E_CONNTIMEDOUT;
			case ECONNREFUSED:
				return E_CONNREFUSED;

			default:
				return status;
		}
	}
}                                 


void SDPCloseServerConnection(SDPServerHandle srvHandle)
{
	closeConnection(srvHandle);
}


/*
 ** Connect to local SDP server on this device
 **
 ** INPUT :
 **
 **   NONE
 **
 ** OUTPUT :
 **
 **   SDPServerHandle** points to a list
 **   of SDP servers found in the piconet
 **
 */
SDPServerHandle SDPOpenLocalServerConnection(void)
{
	struct sockaddr_in	sa;
	int			err;

	if (sdp_flags & SDP_SVC_PROVIDER) {
		sdp_pid = affix_pidof("btsdp", PIDOF_SINGLE, 0);
		if (sdp_pid == 0) {
			/* start server */
			sdp_pid = fork();
			if (sdp_pid < 0) {
				BTERROR("fork() failed\n");
				return -1;
			}
			if (!sdp_pid) {
				/* child */
				err = execlp("btsdp", "btsdp", "-D", NULL);
				if (err)
					BTERROR("Unable to run btsdp\n");
			} else {
				struct sockaddr_in	in;
				int			fd;
				socklen_t		len;
				time_t			stm, tm;

				/* wait for server goes up */
				in.sin_family = AF_INET;
				in.sin_port = SDP_TCP_PORT;
				in.sin_addr.s_addr = inet_addr("127.0.0.1");
				len = sizeof(in);

				fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
				if (fd < 0)
					return -1;
				stm = time(NULL);
				for (;;) {
					err = connect(fd, (struct sockaddr *)&in, len);
					if (!err)
						break;
					tm = time(NULL);
					if (difftime(tm, stm) > 2) {
						close(fd);
						return -1;
					}	
				}
				close(fd);
			}
		}
	}
	/* now try to connect */
	sa.sin_family = PF_INET;
	inet_aton("127.0.0.1", &sa.sin_addr);
	return SDPOpenServerConnection((struct sockaddr_affix*)&sa);
}


/*
 ** Find devices in the piconet.
 ** First arg is a pointer to locations where
 ** discovered devices can be added. "numberOfDevices"
 ** is the max count that pItems can hold
 **
 */
int SDPGeneralInquiry(INQUIRY_ITEM *pItems, int numberOfDevices, 
		int duration, uint8_t *devicesFound)
{
	int	fd, i;
	__u32	bdLap;
	int	status;
	
	INQUIRY_ITEM localItems[32];

	fd = hci_open("bt0");
	if (fd == -1) {
		BTERROR("Cannot open for general inquiry:%s", strerror(errno));
		status = -1;
	} else {
		bdLap = HCI_GIAC;
		status = HCI_Inquiry(fd, duration, 
				32, localItems, devicesFound);
		if (status == 0) {
			INQUIRY_ITEM *ptr;

			ptr = pItems;
			for (i = 0; i < ((numberOfDevices > *devicesFound) ?
						*devicesFound : numberOfDevices); i++) {
				*ptr = localItems[i];
				ptr++;
			}
		}
	}
	close(fd);
	return status;
}

/*
 ** A simple function which returns the time of day in
 ** seconds. Used for updating the service db state
 ** attribute of the service record of the SDP server
 */
long getTimeInSeconds(void)
{
	/*
	 ** To handle failure in gettimeofday, so an old
	 ** value is returned and service does not fail
	 */
	static struct timeval tm;

	gettimeofday(&tm, NULL);

	return tm.tv_sec;
}


/*
 ** Generate unique transaction identifiers
 */
uint16_t generateTransactionId(void)
{
	uint16_t j;

	j=1+(int) (100.0*rand()/(RAND_MAX+1.0));

	return j;
}


/*
 ** Close an existing connection to the SDP server on a said
 ** socket descriptor. 
 */
void closeConnection(int socketFd)
{
	int status;

	if (socketFd != -1) {
		status = close(socketFd);
		if (sdp_mode == SDP_SERVER)
			removeCachedResponse(socketFd, NULL);
	}
}                         


/*
 ** Initialize the SDP machinery
 */
int sdp_init(OperationMode mode)
{
	int status = 0;
	sdp_mode = mode;
	return status;  
}


/*
 ** This must be called to shut the system down
 */
void sdp_shutdown(void)
{
}



/*
 ** A generic send request, wait for response method. Uses
 ** select system call to block for response or timeout
 */
int sendRequestWaitForResponse(
		SDPServerHandle srvHandle,
		char *requestBuffer,
		char *responseBuffer,
		int requestSize,
		int *responseSize
		)
{
	int status = 0;
	int bytesSent = 0;
	int bytesRecvd = 0;
	SdpPduHeader *pduRequestHeader;
	SdpPduHeader *pduResponseHeader;
	fd_set readFds;
	struct timeval timeout;

	pduRequestHeader = (SdpPduHeader *)requestBuffer;
	pduResponseHeader = (SdpPduHeader *)responseBuffer;

	/*
	 ** Send the request, wait for response and if
	 ** not error, set the appropriate values
	 ** and return
	 */
	status = send(srvHandle, requestBuffer, requestSize, 0);
	if (status == -1) {
		BTERROR("Error sending data:%s", strerror(errno));
		goto exit;
	}
	bytesSent = status;
	if (bytesSent != requestSize) {
		BTDEBUG("Attempt to send : %d sent : %d\n", requestSize, bytesSent);
		status = E_FAILURE;
		goto exit;
	}
	BTDEBUG("Bytes sent to handle : %d is : %d\n", srvHandle, bytesSent);        
	BTDUMP(requestBuffer, bytesSent);

	BTDEBUG("Waiting for response\n");
	timeout.tv_sec = WaitTimeForResponse; 
	timeout.tv_usec = 0;

	FD_ZERO(&readFds);
	FD_SET(srvHandle, &readFds);
	status = select(srvHandle+1, &readFds, NULL, NULL, &timeout);
	if (status < 0)
		goto exit;
	
	if (status == 0) {
		status = E_TIMEOUT;
		BTERROR("Client timed out\n");
		goto exit;
	}

	BTDEBUG("Trying to read\n");
	status = recv(srvHandle, responseBuffer, ResponseBufferSize, 0);
	if (status <= 0)
		goto exit;

	bytesRecvd = status;
	BTDEBUG("Read : %d\n", bytesRecvd);
	BTDUMP(responseBuffer, bytesRecvd);
	/*
	 ** Process the response
	 */
	pduResponseHeader = (SdpPduHeader *)responseBuffer;
	if (pduRequestHeader->transactionId ==
			pduResponseHeader->transactionId) {
		*responseSize = bytesRecvd;
		status = 0;
	} else {
		status = E_FAILURE;
	}
exit:
	return status;
}

union sockaddr_sdp {
	struct sockaddr_affix	bt;
	struct sockaddr_in	in;
};

int getMTUFromFd(int socketFd)
{
	union sockaddr_sdp	caddr;
	socklen_t		len = sizeof(caddr);
	int			mtu;

	getsockname(socketFd, (struct sockaddr*)&caddr, &len);
	if (caddr.bt.family == PF_AFFIX)
		mtu = l2cap_getmtu(socketFd);
	else {
		len = sizeof(int);
		getsockopt(socketFd, SOL_SOCKET, SO_SNDBUF, &mtu, &len);
		BTDEBUG("TCP MTU: %d\n", mtu);
	}
	return mtu;
}

/* ------------------------------------------------------- */

/*
 ** This version is for use by SDP server when creating new
 ** service records
 */
ServiceRecord *createServiceRecord(void)
{
	ServiceRecord *svcRec;

	/*
	 ** Create an empty ServiceRecord, but with 
	 ** ServiceRecordHandle set to 0x0000 which
	 ** is the handle for the SDP server itself.
	 **
	 ** The server main program does the rest of
	 ** setting other attributes.
	 */
	svcRec = (ServiceRecord *)malloc(sizeof(ServiceRecord));
	memset((void *)svcRec, 0, sizeof(ServiceRecord));

	svcRec->serviceRecordHandle = 0xffffffff;
	svcRec->ownerSocket = -1;

	return svcRec;
}



/*
 ** Print the contents of a service record
 */
void printServiceRecord(BD_ADDR bdAddr, ServiceRecordHandle svcRecHandle)
{
}


/*
 ** Free the contents of a service record
 */
void freeServiceRecord(ServiceRecord *svcRec)
{
	if (svcRec) {
		free(svcRec->pduForm.data);
		s_list_destroy(&svcRec->targetPattern);
		freeSDPDataSeq(svcRec->attributeList);
		free(svcRec);
	}
}

